import os
from twilio.rest import Client
import re
from datetime import datetime as datetime
from typing import List
import pandas as pd

from mindsdb.utilities import log
from mindsdb.utilities.config import Config

from mindsdb_sql_parser import ast
from mindsdb.integrations.utilities.date_utils import parse_local_date

from mindsdb.integrations.libs.api_handler import APIHandler, APITable

from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions, project_dataframe, filter_dataframe

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

logger = log.getLogger(__name__)


class WhatsAppMessagesTable(APITable):
    def select(self, query: ast.Select) -> Response:
        """
        Retrieves messages sent/received from the database using Twilio Whatsapp API
        Returns
            Response: conversation_history
        """

        # Extract comparison conditions from the query
        conditions = extract_comparison_conditions(query.where)
        params = {}
        filters = []

        # Build the filters and parameters for the query
        for op, arg1, arg2 in conditions:
            if op == 'or':
                raise NotImplementedError('OR is not supported')

            if arg1 == 'sent_at' and arg2 is not None:
                date = parse_local_date(arg2)
                if op == '>':
                    params['date_sent_after'] = date
                elif op == '<':
                    params['date_sent_before'] = date
                else:
                    raise NotImplementedError

                # also add to post query filter because date_sent_after=date1 will include date1
                filters.append([op, arg1, arg2])

            elif arg1 == 'sid':
                if op == '=':
                    params['sid'] = arg2
                else:
                    NotImplementedError('Only  "from_number=" is implemented')

            elif arg1 == 'from_number':
                if op == '=':
                    params['from_number'] = arg2
                else:
                    NotImplementedError('Only  "from_number=" is implemented')

            elif arg1 == 'to_number':
                if op == '=':
                    params['to_number'] = arg2
                else:
                    NotImplementedError('Only  "to_number=" is implemented')

            else:
                filters.append([op, arg1, arg2])

        # Fetch messages based on the filters
        result = self.handler.fetch_messages(params, df=True)

        # filter targets
        result = filter_dataframe(result, filters)

        # If limit is specified
        if query.limit is not None:
            result = result[:int(query.limit.value)]

        # project targets
        result = project_dataframe(result, query.targets, self.get_columns())

        return result

    def get_columns(self):
        return [
            'sid',
            'from_number',
            'to_number',
            'body',
            'direction',
            'msg_status',
            'sent_at',  # datetime.strptime(str(msg.date_sent), '%Y-%m-%d %H:%M:%S%z'),
            'account_sid',
            'price',
            'price_unit',
            'api_version',
            'uri'
        ]

    def insert(self, query: ast.Insert):
        """
        Sends a whatsapp message

        Args:
            body: message body
            from_number: number from which to send the message
            to_number: number to which message will be sent
        """

        # get column names and values from the query
        columns = [col.name for col in query.columns]

        ret = []

        insert_params = ["body", "from_number", "to_number"]
        for row in query.values:
            params = dict(zip(columns, row))

            # Check text length
            max_text_len = 1500
            text = params["body"]
            words = re.split('( )', text)
            messages = []

            """
                Regex for matching if any URls are present, if yes then replace with string of hyphens(-)

                Example:
                    words = ['Check', ' ', 'out', ' ', 'this', ' ', 'cool', ' ', 'website:', ' ', 'https://example.com.', "It's", ' ', 'awesome!']

                    After parsing through regex ('https://example.com') URL is matched

                    Final output:
                    messages = ['Check - out - this - cool - website: ----------------------- "It\'s - awesome!']
            """

            text2 = ''
            pattern = r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'
            for word in words:
                # replace the links in word to string with the length as twitter short url (23)
                word2 = re.sub(pattern, '-' * 23, word)
                if len(text2) + len(word2) > max_text_len - 3 - 7:  # 3 is for ..., 7 is for (10/11)
                    messages.append(text2.strip())

                    text2 = ''
                text2 += word

            # Parse last message
            if text2.strip() != '':
                messages.append(text2.strip())

            len_messages = len(messages)

            # Modify message based on the length
            for i, text in enumerate(messages):
                if i < len_messages - 1:
                    text += '...'
                else:
                    text += ' '

                if i >= 1:
                    text += f'({i + 1}/{len_messages})'

                # Pass parameters and call 'send_message'
                params['body'] = text
                params_to_send = {key: params[key] for key in insert_params if (key in params)}
                ret_row = self.handler.send_message(params_to_send, ret_as_dict=True)

                # Save the results
                ret_row['body'] = text
                ret.append(ret_row)

        return pd.DataFrame(ret)


class WhatsAppHandler(APIHandler):
    """
    A class for handling connections and interactions with Twilio WhatsApp API.
    Args:
        account_sid(str): Accound ID of the twilio account.
        auth_token(str): Authentication Token obtained from the twilio account.
    """

    def __init__(self, name=None, **kwargs):
        """
        Initializes the connection by checking all the params are provided by the user.
        """
        super().__init__(name)

        args = kwargs.get('connection_data', {})
        self.connection_args = {}
        handler_config = Config().get('whatsapp_handler', {})
        for k in ['account_sid', 'auth_token']:
            if k in args:
                self.connection_args[k] = args[k]
            elif f'TWILIO_{k.upper()}' in os.environ:
                self.connection_args[k] = os.environ[f'TWILIO_{k.upper()}']
            elif k in handler_config:
                self.connection_args[k] = handler_config[k]
        self.client = None
        self.is_connected = False

        messages = WhatsAppMessagesTable(self)
        self._register_table('messages', messages)

    def connect(self):
        """
        Authenticate with the Twilio API using the provided `account_SID` and `auth_token`.
        """
        if self.is_connected is True:
            return self.client

        self.client = Client(
            self.connection_args['account_sid'],
            self.connection_args['auth_token']
        )

        self.is_connected = True
        return self.client

    def check_connection(self) -> StatusResponse:
        """
        Checks the connection by performing a basic operation with the Twilio API.
        """
        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True

        except Exception as e:
            response.error_message = f'Error connecting to Twilio API: {str(e)}. Check credentials.'
            logger.error(response.error_message)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def parse_native_query(self, query_string: str):
        """Parses the native query string of format method(arg1=val1, arg2=val2, ...) and returns the method name and arguments."""

        # Adjust regex to account for the possibility of no arguments inside the parenthesis
        match = re.match(r'(\w+)\(([^)]*)\)', query_string)
        if not match:
            raise ValueError(f"Invalid query format: {query_string}")

        method_name = match.group(1)
        arg_string = match.group(2)

        # Extract individual arguments
        args = {}
        if arg_string:  # Check if there are any arguments
            for arg in arg_string.split(','):
                arg = arg.strip()
                key, value = arg.split('=')
                args[key.strip()] = value.strip()

        return method_name, args

    def native_query(self, query_string: str = None):
        """
        Retreievs the native query from the `parse_native_query` and calls appropriate function and returns the result of the query as a Response object.
        """
        method_name, params = self.parse_native_query(query_string)
        if method_name == 'send_message':
            response = self.send_message(params)
        else:
            raise ValueError(f"Method '{method_name}' not supported by TwilioHandler")

        return response

    def fetch_messages(self, params, df=False):
        """
        Gets conversation history

        Returns:
            Response: conversation history
        """
        limit = int(params.get('limit', 1000))
        sid = params.get('sid', None)
        # Convert date strings to datetime objects if provided
        date_sent_after = params.get('date_sent_after', None)
        date_sent_before = params.get('date_sent_before', None)
        # Extract 'from_' and 'body' search criteria from params
        from_number = params.get('from_number', None)
        to_number = params.get('to_number', None)
        args = {
            'limit': limit,
            'date_sent_after': date_sent_after,
            'date_sent_before': date_sent_before,
            'from_': from_number,
            'to': to_number
        }

        args = {arg: val for arg, val in args.items() if val is not None}
        if sid:
            messages = [self.client.messages(sid).fetch()]
        else:
            messages = self.client.messages.list(**args)

        # Extract all possible properties for each message
        data = []
        for msg in messages:
            msg_data = {
                'sid': msg.sid,
                'to_number': msg.to,
                'from_number': msg.from_,
                'body': msg.body,
                'direction': msg.direction,
                'msg_status': msg.status,
                'sent_at': msg.date_created.replace(tzinfo=None),
                'account_sid': msg.account_sid,
                'price': msg.price,
                'price_unit': msg.price_unit,
                'api_version': msg.api_version,
                'uri': msg.uri,
                # 'media_url': [media.uri for media in msg.media.list()]
                # ... Add other properties as needed
            }
            data.append(msg_data)

        # Create a DataFrame
            result_df = pd.DataFrame(data)

            # Filter rows where 'from_number' or 'to_number' begins with 'whatsapp:'
            result_df = result_df[result_df['from_number'].str.startswith('whatsapp:') | result_df['to_number'].str.startswith('whatsapp:')]

        if df is True:
            return pd.DataFrame(result_df)
        return Response(RESPONSE_TYPE.TABLE, data_frame=pd.DataFrame(result_df))

    def send_message(self, params, ret_as_dict=False) -> Response:
        """
        Sends a message to the given Whatsapp number.

        Args:
            body: message body
            from_number: number from which to send the message
            to_number: number to which message will be sent
        """
        try:
            message = self.client.messages.create(
                body=params.get('body'),
                to=params.get('to_number'),
                from_=params.get('from_number')
            )

            if ret_as_dict is True:
                return {"sid": message.sid, "from": message.from_, "to": message.to, "message": message.body, "status": message.status}

            return Response(
                RESPONSE_TYPE.MESSAGE,
                sid=message.sid,
                from_=message.from_,
                to=message.to,
                body=message.body,
                status=message.status
            )

        except Exception as e:
            # Log the exception for debugging purposes
            logger.error(f"Error sending message: {str(e)}")
            logger.exception(f"Error sending message: {str(e)}")
            raise Exception("Error sending message")

    def call_whatsapp_api(self, method_name: str = None, params: dict = None):
        """
        Calls specific method specified.

        Args:
            method_name: to call specific method
            params: parameters to call the method

        Returns:
            List of dictionaries as a result of the method call
        """
        api = self.connect()
        method = getattr(api, method_name)

        try:
            result = method(**params)
        except Exception as e:
            error = f"Error calling method '{method_name}' with params '{params}': {e}"
            logger.error(error)
            raise e

        if 'messages' in result:
            result['messages'] = self.convert_channel_data(result['messages'])

        return [result]

    def convert_channel_data(self, messages: List[dict]):
        """
        Convert the list of channel dictionaries to a format that can be easily used in the data pipeline.

        Args:
            channels: A list of channel dictionaries.

        Returns:
            A list of channel dictionaries with modified keys and values.
        """
        new_messages = []
        for message in messages:
            new_message = {
                'id': message['id'],
                'name': message['name'],
                'created': datetime.fromtimestamp(float(message['created']))
            }
            new_messages.append(new_message)
        return new_messages



from contextlib import contextmanager

import json
import duckdb
import pandas as pd
import fsspec
import google.auth
from google.cloud import storage
from typing import Text, Dict, Optional, List
from duckdb import DuckDBPyConnection

from mindsdb.integrations.handlers.gcs_handler.gcs_tables import (
    ListFilesTable,
    FileTable
)
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb_sql_parser.ast import Select, Identifier, Insert, Star, Constant

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

from mindsdb.integrations.libs.api_handler import APIHandler

logger = log.getLogger(__name__)


class GcsHandler(APIHandler):
    """
    This handler handles connection and execution of the SQL statements on GCS.
    """

    name = 'gcs'

    supported_file_formats = ['csv', 'tsv', 'json', 'parquet']

    def __init__(self, name: Text, connection_data: Optional[Dict], **kwargs):
        """
        Initializes the handler.

        Args:
            name (Text): The name of the handler instance.
            connection_data (Dict): The connection data required to connect to the GCS account.
            kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.is_select_query = False
        self.service_account_json = None
        self.connection = None

        if 'service_account_keys' not in self.connection_data and 'service_account_json' not in self.connection_data:
            raise ValueError('service_account_keys or service_account_json parameter must be provided.')

        if 'service_account_json' in self.connection_data:
            self.service_account_json = self.connection_data["service_account_json"]

        if 'service_account_keys' in self.connection_data:
            with open(self.connection_data["service_account_keys"], "r") as f:
                self.service_account_json = json.loads(f.read())

        self.is_connected = False

        self._files_table = ListFilesTable(self)

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> DuckDBPyConnection:
        """
        Establishes a connection to the GCS account via DuckDB.

        Raises:
            KeyError: If the required connection parameters are not provided.

        Returns:
            DuckDBPyConnection : A client object to the GCS account.
        """
        if self.is_connected is True:
            return self.connection

        # Connect to GCS and configure mandatory credentials.
        self.connection = self._connect_storage_client()
        self.is_connected = True

        return self.connection

    @contextmanager
    def _connect_duckdb(self):
        """
        Creates temporal duckdb database which is able to connect to the GCS account.
        Have to be used as context manager

        Returns:
            DuckDBPyConnection
        """
        # Connect to GCS via DuckDB.
        duckdb_conn = duckdb.connect(":memory:")

        # Configure mandatory credentials.
        credentials, project_id = google.auth.load_credentials_from_dict(self.service_account_json)
        gcs = fsspec.filesystem("gcs", project=project_id, credentials=credentials)
        duckdb_conn = duckdb.connect()
        duckdb_conn.register_filesystem(gcs)

        try:
            yield duckdb_conn
        finally:
            duckdb_conn.close()

    def _connect_storage_client(self) -> storage.Client:
        """
        Establishes a connection to the GCS account via google-cloud-storage.

        Returns:
            storage.Client: A client object to the GCS account.
        """
        return storage.Client.from_service_account_info(self.service_account_json)

    def disconnect(self):
        """
        Closes the connection to the GCP account if it's currently open.
        """
        if not self.is_connected:
            return
        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the GCS bucket.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        # Check connection via storage client.
        try:
            storage_client = self._connect_storage_client()
            if 'bucket' in self.connection_data:
                storage_client.get_bucket(self.connection_data['bucket'])
            else:
                storage_client.list_buckets()
            response.success = True
            storage_client.close()
        except Exception as e:
            logger.error(f'Error connecting to GCS with the given credentials, {e}!')
            response.error_message = str(e)

        if response.success and need_to_close:
            self.disconnect()

        elif not response.success and self.is_connected:
            self.is_connected = False

        return response

    def _get_bucket(self, key):
        if 'bucket' in self.connection_data:
            return self.connection_data['bucket'], key

        # get bucket from first part of the key
        ar = key.split('/')
        return ar[0], '/'.join(ar[1:])

    def read_as_table(self, key) -> pd.DataFrame:
        """
        Read object as dataframe. Uses duckdb
        """

        bucket, key = self._get_bucket(key)

        with self._connect_duckdb() as connection:

            cursor = connection.execute(f"SELECT * FROM 'gs://{bucket}/{key}'")

            return cursor.fetchdf()

    def _read_as_content(self, key) -> None:
        """
        Read object as content
        """
        bucket, key = self._get_bucket(key)

        client = self.connect()

        bucket = client.bucket(bucket)
        blob = bucket.blob(key)
        return blob.download_as_string()

    def add_data_to_table(self, key, df) -> None:
        """
        Writes the table to a file in the gcs bucket.

        Raises:
            CatalogException: If the table does not exist in the DuckDB connection.
        """

        # Check if the file exists in the gcs bucket.
        bucket, key = self._get_bucket(key)

        storage_client = self._connect_storage_client()
        bucketObj = storage_client.bucket(bucket)
        stats = storage.Blob(bucket=bucketObj, name=key).exists(storage_client)
        storage_client.close()
        if not stats:
            raise Exception(f'Error querying the file {key} in the bucket {bucket}!')

        with self._connect_duckdb() as connection:
            # copy
            connection.execute(f"CREATE TABLE tmp_table AS SELECT * FROM 'gs://{bucket}/{key}'")

            # insert
            connection.execute("INSERT INTO tmp_table BY NAME SELECT * FROM df")

            # upload
            connection.execute(f"COPY tmp_table TO 'gs://{bucket}/{key}'")

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Raises:
            ValueError: If the file format is not supported or the file does not exist in the GCS bucket.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """

        self.connect()

        if isinstance(query, Select):
            table_name = query.from_table.parts[-1]

            if table_name == 'files':
                table = self._files_table
                df = table.select(query)

                # add content
                has_content = False
                for target in query.targets:
                    if isinstance(target, Identifier) and target.parts[-1].lower() == 'content':
                        has_content = True
                        break
                if has_content:
                    df['content'] = df['path'].apply(self._read_as_content)
            else:
                extension = table_name.split('.')[-1]
                if extension not in self.supported_file_formats:
                    logger.error(f'The file format {extension} is not supported!')
                    raise ValueError(f'The file format {extension} is not supported!')

                table = FileTable(self, table_name=table_name)
                df = table.select(query)

            response = Response(
                RESPONSE_TYPE.TABLE,
                data_frame=df
            )
        elif isinstance(query, Insert):
            table_name = query.table.parts[-1]
            table = FileTable(self, table_name=table_name)
            table.insert(query)
            response = Response(RESPONSE_TYPE.OK)
        else:
            raise NotImplementedError

        return response

    def get_objects(self, limit=None, buckets=None) -> List[dict]:
        storage_client = self._connect_storage_client()
        if "bucket" in self.connection_data:
            add_bucket_to_name = False
            scan_buckets = [self.connection_data["bucket"]]
        else:
            add_bucket_to_name = True
            scan_buckets = [b.name for b in storage_client.list_buckets()]

        objects = []
        for bucket in scan_buckets:
            if buckets is not None and bucket not in buckets:
                continue

            blobs = storage_client.list_blobs(bucket)
            if not blobs:
                continue

            for blob in blobs:
                if blob.storage_class != 'STANDARD':
                    continue

                obj = {}
                obj['Bucket'] = bucket
                if add_bucket_to_name:
                    # bucket is part of the name
                    obj['Key'] = f'{bucket}/{blob.name}'
                objects.append(obj)
            if limit is not None and len(objects) >= limit:
                break

        return objects

    def get_tables(self) -> Response:
        """
        Retrieves a list of tables (objects) in the gcs bucket.

        Each object is considered a table. Only the supported file formats are considered as tables.

        Returns:
            Response: A response object containing the list of tables and views, formatted as per the `Response` class.
        """

        # Get only the supported file formats.
        # Wrap the object names with backticks to prevent SQL syntax errors.
        supported_names = [
            f"`{obj['Key']}`"
            for obj in self.get_objects()
            if obj['Key'].split('.')[-1] in self.supported_file_formats
        ]

        # virtual table with list of files
        supported_names.insert(0, 'files')

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                supported_names,
                columns=['table_name']
            )
        )

        return response

    def get_columns(self, table_name: str) -> Response:
        """
        Retrieves column details for a specified table (object) in the gcs bucket.

        Args:
            table_name (Text): The name of the table for which to retrieve column information.

        Raises:
            ValueError: If the 'table_name' is not a valid string.

        Returns:
            Response: A response object containing the column details, formatted as per the `Response` class.
        """
        query = Select(
            targets=[Star()],
            from_table=Identifier(parts=[table_name]),
            limit=Constant(1)
        )

        result = self.query(query)

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                {
                    'column_name': result.data_frame.columns,
                    'data_type': [data_type if data_type != 'object' else 'string' for data_type in result.data_frame.dtypes]
                }
            )
        )

        return response



from flask import request
from flask_restx import Resource

from mindsdb.api.http.namespaces.configs.webhooks import ns_conf
from mindsdb.interfaces.chatbot.chatbot_controller import ChatBotController
from mindsdb.metrics.metrics import api_endpoint_metrics


# Stores the memory of the various chat-bots mapped by their webhook tokens.
# This is required because each time a new request is made, a new instance of the ChatBotTask is created.
# This causes the memory to be lost.
chat_bot_memory = {}


@ns_conf.route('/chatbots/<webhook_token>')
class ChatbotWebhooks(Resource):
    @ns_conf.doc('chatbots_webhook')
    @api_endpoint_metrics('POST', '/webhooks/chatbots/<webhook_token>')
    def post(self, webhook_token: str) -> None:
        """
        This endpoint is used to receive messages posted by bots from different platforms.

        Args:
        webhook_token (str): The token of the webhook. It is used to uniquely identify the webhook.
        """
        request_data = request.json

        chat_bot_controller = ChatBotController()
        return chat_bot_controller.on_webhook(webhook_token, request_data, chat_bot_memory)



from typing import Optional

import pandas as pd

from mindsdb.integrations.handlers.sentence_transformers_handler.settings import Parameters

from mindsdb.integrations.handlers.rag_handler.settings import load_embeddings_model, df_to_documents


from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class SentenceTransformersHandler(BaseMLEngine):
    name = "sentence transformers"

    def __init__(self, model_storage, engine_storage, **kwargs) -> None:
        super().__init__(model_storage, engine_storage, **kwargs)
        self.generative = True

    def create(self, target, df=None, args=None, **kwargs):
        """creates embeddings model and persists"""

        args = args["using"]

        valid_args = Parameters(**args)
        self.model_storage.json_set("args", valid_args.model_dump())

    def predict(self, df, args=None):
        """loads persisted embeddings model and gets embeddings on input text column(s)"""

        args = self.model_storage.json_get("args")

        if isinstance(df['content'].iloc[0], list) and len(df['content']) == 1:
            # allow user to pass in a list of strings in where clause
            # i.e where content = ['hello', 'world'] or where content = (select content from some_db.some_table)
            input_df = df.copy()
            df = pd.DataFrame(data={"content": input_df['content'].iloc[0]})

        # get text columns if specified
        if isinstance(args['text_columns'], str):
            columns = [args['text_columns']]

        elif isinstance(args['text_columns'], list):
            columns = args['text_columns']

        elif args['text_columns'] is None:
            # assume all columns are text columns
            logger.info("No text columns specified, assuming all columns are text columns")
            columns = df.columns.tolist()

        else:
            raise ValueError(f"Invalid value for text_columns: {args['text_columns']}")

        documents = df_to_documents(df=df, page_content_columns=columns)

        content = [doc.page_content for doc in documents]
        metadata = [doc.metadata for doc in documents]

        model = load_embeddings_model(args['embeddings_model_name'])

        embeddings = model.embed_documents(texts=content)

        embeddings_df = pd.DataFrame(data={"content": content, "embeddings": embeddings, "metadata": metadata})

        return embeddings_df

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:

        args = self.model_storage.json_get("args")

        if attribute == "args":
            return pd.DataFrame(args.items(), columns=["key", "value"])



from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.integrations.handlers.financial_modeling_prep_handler.financial_modeling_tables import HistoricalPriceTable

from mindsdb.utilities import log
import requests

logger = log.getLogger(__name__)


class FinancialModelingHandler(APIHandler):

    name = "financial_modeling_prep"

    def __init__(self, name, connection_data: dict, **kwargs):
        super().__init__(name)

        self.api_key = None
        self.connection_data = connection_data
        if "api_key" not in connection_data:
            raise Exception(
                "FINANCIAL_MODELING engine requires an API key. Retrieve an API key from https://site.financialmodelingprep.com/developer. See financial_modeling_prep_handler/README.MD on how to include API key in query."
            )
        self.api_key = connection_data['api_key']
        self.client = None
        self.is_connected = False

        historical_prices = HistoricalPriceTable(self)
        self._register_table('historical_prices', historical_prices)

    def connect(self):
        self.is_connected = True
        base_url = "https://financialmodelingprep.com/api/v3/historical-price-full/"
        return base_url

    def check_connection(self) -> StatusResponse:
        """ Check connection to the handler
        Returns:
            HandlerStatusResponse
        """
        base_url = 'https://financialmodelingprep.com/api/v3/search'
        param = {
            'query': 'AA',
            'apikey': self.api_key,
            'limit': 5
        }

        response = requests.get(base_url, param)
        if response.status_code == 200:
            return StatusResponse(success=True)
        else:
            raise Exception(
                "API key provided in query is not valid. Retrieve a valid API key from https://site.financialmodelingprep.com/developer. See financial_modeling_prep_handler/README.MD on how to include API key in query."
            )



import pandas as pd

from typing import List

from mindsdb.integrations.libs.api_handler import APITable

from mindsdb_sql_parser import ast

from mindsdb.integrations.utilities.handlers.query_utilities import SELECTQueryParser, SELECTQueryExecutor

from mindsdb.utilities import log

logger = log.getLogger(__name__)


class PagesTable(APITable):
    """The MediaWiki Pages Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls MediaWiki pages data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Sendinblue Email Campaigns matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'pages',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        title, page_id = None, None
        for condition in where_conditions:
            if condition[1] == 'title':
                if condition[0] != '=':
                    raise ValueError(f"Unsupported operator '{condition[0]}' for column '{condition[1]}' in WHERE clause.")
                title = condition[2]
            elif condition[1] == 'pageid':
                if condition[0] != '=':
                    raise ValueError(f"Unsupported operator '{condition[0]}' for column '{condition[1]}' in WHERE clause.")
                page_id = condition[2]
            else:
                raise ValueError(f"Unsupported column '{condition[1]}' in WHERE clause.")

        pages_df = pd.json_normalize(self.get_pages(title=title, page_id=page_id, limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            pages_df,
            selected_columns,
            [],
            order_by_conditions
        )
        pages_df = select_statement_executor.execute_query()

        return pages_df

    def get_columns(self) -> List[str]:
        return ['pageid', 'title', 'original_title', 'content', 'summary', 'url', 'categories']

    def get_pages(self, title: str = None, page_id: int = None, limit: int = 20):
        query_parts = []

        query_parts.append(f'intitle:{title}') if title is not None else None
        query_parts.append(f'pageid:{page_id}') if page_id is not None else None

        search_query = ' | '.join(query_parts)

        connection = self.handler.connect()

        if search_query:
            return [self.convert_page_to_dict(connection.page(result, auto_suggest=False)) for result in connection.search(search_query, results=limit)]
        else:
            return [self.convert_page_to_dict(connection.page(result, auto_suggest=False)) for result in connection.random(pages=limit)]

    def convert_page_to_dict(self, page):
        result = {}
        attributes = self.get_columns()

        for attribute in attributes:
            try:
                result[attribute] = getattr(page, attribute)
            except KeyError:
                logger.debug(f"Error accessing '{attribute}' attribute. Skipping...")

        return result


"""training_progress

Revision ID: 3154382dab17
Revises: ee63d868fa84
Create Date: 2023-02-25 15:12:02.828938

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3154382dab17'
down_revision = 'ee63d868fa84'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('training_phase_current', sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column('training_phase_total', sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column('training_phase_name', sa.String(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_column('training_phase_name')
        batch_op.drop_column('training_phase_total')
        batch_op.drop_column('training_phase_current')

    # ### end Alembic commands ###



from typing import List
import copy
from dataclasses import dataclass, field

from mindsdb_sql_parser import ast
from mindsdb_sql_parser.ast import (
    Select, Identifier, BetweenOperation, Join, Star, BinaryOperation, Constant,
    NativeQuery, Parameter
)

from mindsdb.integrations.utilities.query_traversal import query_traversal

from mindsdb.api.executor.planner.exceptions import PlanningException
from mindsdb.api.executor.planner.steps import (
    FetchDataframeStep, JoinStep, ApplyPredictorStep, SubSelectStep, QueryStep, MapReduceStep
)
from mindsdb.api.executor.planner.utils import filters_to_bin_op
from mindsdb.api.executor.planner.plan_join_ts import PlanJoinTSPredictorQuery


@dataclass
class TableInfo:
    integration: str
    table: Identifier
    aliases: List[str] = field(default_factory=List)
    conditions: List = None
    sub_select: ast.ASTNode = None
    predictor_info: dict = None
    join_condition = None
    index: int = None


class PlanJoin:

    def __init__(self, planner):
        self.planner = planner

    def is_timeseries(self, query):

        join = query.from_table
        l_predictor = self.planner.get_predictor(join.left) if isinstance(join.left, Identifier) else None
        r_predictor = self.planner.get_predictor(join.right) if isinstance(join.right, Identifier) else None
        if l_predictor and l_predictor.get('timeseries'):
            return True
        if r_predictor and r_predictor.get('timeseries'):
            return True

    def check_single_integration(self, query):
        query_info = self.planner.get_query_info(query)

        # can we send all query to integration?

        # one integration and not mindsdb objects in query
        if (
                len(query_info['mdb_entities']) == 0
                and len(query_info['integrations']) == 1
                and 'files' not in query_info['integrations']
                and 'views' not in query_info['integrations']
        ):

            int_name = list(query_info['integrations'])[0]
            # if is sql database
            if self.planner.integrations.get(int_name, {}).get('class_type') != 'api':

                # send to this integration
                return int_name
        return None

    def plan(self, query, integration=None):
        # FIXME: Tableau workaround, INFORMATION_SCHEMA with Where
        # if isinstance(join.right, Identifier) \
        #         and self.resolve_database_table(join.right)[0] == 'INFORMATION_SCHEMA':
        #     pass

        # send join to integration as is?
        integration_to_send = self.check_single_integration(query)
        if integration_to_send:
            self.planner.prepare_integration_select(integration_to_send, query)

            last_step = self.planner.plan.add_step(FetchDataframeStep(integration=integration_to_send, query=query))
            return last_step
        elif self.is_timeseries(query):
            return PlanJoinTSPredictorQuery(self.planner).plan(query, integration)
        else:
            return PlanJoinTablesQuery(self.planner).plan(query)


class PlanJoinTablesQuery:

    def __init__(self, planner):
        self.planner = planner

        # index to lookup tables
        self.tables_idx = None
        self.tables = []
        self.tables_fetch_step = {}

        self.step_stack = None
        self.query_context = {}

        self.partition = None

    def plan(self, query):
        self.tables_idx = {}
        join_step = self.plan_join_tables(query)

        if (
                query.group_by is not None
                or query.order_by is not None
                or query.having is not None
                or query.distinct is True
                or query.where is not None
                or query.limit is not None
                or query.offset is not None
                or len(query.targets) != 1
                or not isinstance(query.targets[0], Star)
        ):
            query2 = copy.deepcopy(query)
            query2.from_table = None
            query2.using = None
            query2.cte = None
            sup_select = QueryStep(query2, from_table=join_step.result)
            self.planner.plan.add_step(sup_select)
            return sup_select
        return join_step

    def resolve_table(self, table):
        # gets integration for table and name to access to it
        table = copy.deepcopy(table)
        # get possible table aliases
        aliases = []
        if table.alias is not None:
            # to lowercase
            parts = tuple(map(str.lower, table.alias.parts))
            aliases.append(parts)
        else:
            for i in range(0, len(table.parts)):
                parts = table.parts[i:]
                parts = tuple(map(str.lower, parts))
                aliases.append(parts)

        # try to use default namespace
        integration = self.planner.default_namespace
        if len(table.parts) > 0:
            if table.parts[0] in self.planner.databases:
                integration = table.parts.pop(0)
            else:
                integration = self.planner.default_namespace

        if integration is None and not hasattr(table, 'sub_select'):
            raise PlanningException(f'Integration not found for: {table}')

        sub_select = getattr(table, 'sub_select', None)

        return TableInfo(integration, table, aliases, conditions=[], sub_select=sub_select)

    def get_table_for_column(self, column: Identifier):
        if not isinstance(column, Identifier):
            return
        # to lowercase
        parts = tuple(map(str.lower, column.parts[:-1]))
        if parts in self.tables_idx:
            return self.tables_idx[parts]

    def get_join_sequence(self, node, condition=None):
        sequence = []
        if isinstance(node, Identifier):
            # resolve identifier

            table_info = self.resolve_table(node)
            for alias in table_info.aliases:
                self.tables_idx[alias] = table_info

            table_info.index = len(self.tables)
            self.tables.append(table_info)

            table_info.predictor_info = self.planner.get_predictor(node)

            if condition is not None:
                table_info.join_condition = condition
            sequence.append(table_info)

        elif isinstance(node, Join):
            # create sequence: 1)table1, 2)table2, 3)join 1 2, 4)table 3, 5)join 3 4

            # put all tables before
            sequence2 = self.get_join_sequence(node.left)
            for item in sequence2:
                sequence.append(item)

            sequence2 = self.get_join_sequence(node.right, condition=node.condition)
            if len(sequence2) != 1:
                raise PlanningException('Unexpected join nesting behavior')

            # put next table
            sequence.append(sequence2[0])

            # put join
            sequence.append(node)

        else:
            raise NotImplementedError()
        return sequence

    def check_node_condition(self, node):

        col_idx = 0
        if len(node.args) == 2:
            if not isinstance(node.args[col_idx], Identifier):
                # try to use second arg, could be: 'x'=col
                col_idx = 1

        # check the case col <condition> constant, col between constant and constant
        for i, arg in enumerate(node.args):
            if i == col_idx:
                if not isinstance(arg, Identifier):
                    return
            else:
                if not isinstance(arg, (Constant, Parameter)):
                    return

        # checked, find table and store condition

        node2 = copy.deepcopy(node)

        arg1 = node2.args[col_idx]

        if len(arg1.parts) < 2:
            return

        table_info = self.get_table_for_column(arg1)
        if table_info is None:
            raise PlanningException(f'Table not found for identifier: {arg1.to_string()}')

        # keep only column name
        arg1.parts = [arg1.parts[-1]]

        node2._orig_node = node
        table_info.conditions.append(node2)

    def check_query_conditions(self, query):
        # get conditions for tables
        binary_ops = []

        def _check_node_condition(node, **kwargs):
            if isinstance(node, BetweenOperation):
                self.check_node_condition(node)

            if isinstance(node, BinaryOperation):
                binary_ops.append(node.op)

                self.check_node_condition(node)

        query_traversal(query.where, _check_node_condition)

        self.query_context['binary_ops'] = binary_ops

    def check_use_limit(self, query_in, join_sequence):
        # use limit for first table?
        # if only models
        use_limit = False
        if query_in.having is None or query_in.group_by is None and query_in.limit is not None:

            join = None
            use_limit = True
            for item in join_sequence:
                if isinstance(item, TableInfo):
                    if item.predictor_info is None and item.sub_select is None:
                        if join is not None:
                            if join.join_type.upper() != 'LEFT JOIN':
                                use_limit = False
                elif isinstance(item, Join):
                    join = item
        self.query_context['use_limit'] = use_limit

    def plan_join_tables(self, query_in):

        # plan all nested selects in 'where'
        find_selects = self.planner.get_nested_selects_plan_fnc(self.planner.default_namespace, force=True)
        query_in.targets = query_traversal(query_in.targets, find_selects)
        query_traversal(query_in.where, find_selects)

        query = copy.deepcopy(query_in)

        # replace sub selects, with identifiers with links to original selects
        def replace_subselects(node, **args):
            if isinstance(node, Select) or isinstance(node, NativeQuery) or isinstance(node, ast.Data):
                name = f't_{id(node)}'
                node2 = Identifier(name, alias=node.alias)

                # save in attribute
                if isinstance(node, NativeQuery) or isinstance(node, ast.Data):
                    # wrap to select
                    node = Select(targets=[Star()], from_table=node)
                node2.sub_select = node
                return node2

        query_traversal(query.from_table, replace_subselects)

        # get all join tables, form join sequence
        join_sequence = self.get_join_sequence(query.from_table)

        # find tables for identifiers used in query
        def _check_identifiers(node, is_table, **kwargs):
            if not is_table and isinstance(node, Identifier):
                if len(node.parts) > 1:
                    table_info = self.get_table_for_column(node)
                    if table_info is None:
                        raise PlanningException(f'Table not found for identifier: {node.to_string()}')

                    # # replace identifies name
                    col_parts = list(table_info.aliases[-1])
                    col_parts.append(node.parts[-1])
                    node.parts = col_parts

        query_traversal(query, _check_identifiers)

        self.check_query_conditions(query)

        # workaround for 'model join table': swap tables:
        if len(join_sequence) == 3 and join_sequence[0].predictor_info is not None:
            join_sequence = [join_sequence[1], join_sequence[0], join_sequence[2]]

        self.check_use_limit(query_in, join_sequence)

        # create plan
        # TODO add optimization: one integration without predictor

        self.step_stack = []
        for item in join_sequence:
            if isinstance(item, TableInfo):

                if item.sub_select is not None:
                    self.process_subselect(item)
                elif item.predictor_info is not None:
                    self.process_predictor(item, query_in)
                else:
                    # is table
                    self.process_table(item, query_in)

            elif isinstance(item, Join):
                step_right = self.step_stack.pop()
                step_left = self.step_stack.pop()

                new_join = copy.deepcopy(item)

                # TODO
                new_join.left = Identifier('tab1')
                new_join.right = Identifier('tab2')
                new_join.implicit = False

                step = self.add_plan_step(JoinStep(left=step_left.result, right=step_right.result, query=new_join))

                self.step_stack.append(step)

        query_in.where = query.where

        self.close_partition()
        return self.step_stack.pop()

    def process_subselect(self, item):
        # is sub select
        item.sub_select.alias = None
        item.sub_select.parentheses = False
        step = self.planner.plan_select(item.sub_select)

        where = filters_to_bin_op(item.conditions)

        # apply table alias
        query2 = Select(targets=[Star()], where=where)
        if item.table.alias is None:
            raise PlanningException(f'Subselect in join have to be aliased: {item.sub_select.to_string()}')
        table_name = item.table.alias.parts[-1]

        add_absent_cols = False
        if hasattr(item.sub_select, 'from_table') and \
                isinstance(item.sub_select.from_table, ast.Data):
            add_absent_cols = True

        step2 = SubSelectStep(query2, step.result, table_name=table_name, add_absent_cols=add_absent_cols)
        step2 = self.add_plan_step(step2)
        self.step_stack.append(step2)

    def process_table(self, item, query_in):
        table = copy.deepcopy(item.table)
        table.parts.insert(0, item.integration)
        query2 = Select(from_table=table, targets=[Star()])
        # parts = tuple(map(str.lower, table_name.parts))
        conditions = item.conditions
        if 'or' in self.query_context['binary_ops']:
            # not use conditions
            conditions = []

        conditions += self.get_filters_from_join_conditions(item)

        if self.query_context['use_limit']:
            order_by = None
            if query_in.order_by is not None:
                order_by = []
                # all order column be from this table
                for col in query_in.order_by:
                    table_info = self.get_table_for_column(col.field)
                    if table_info is None or table_info.table != item.table:
                        order_by = False
                        break
                    col = copy.deepcopy(col)
                    col.field.parts = [col.field.parts[-1]]
                    order_by.append(col)

            if order_by is not False:
                # copy limit from upper query
                query2.limit = query_in.limit
                # move offset from upper query
                query2.offset = query_in.offset
                query_in.offset = None
                # copy order
                query2.order_by = order_by

            self.query_context['use_limit'] = False
        for cond in conditions:
            if query2.where is not None:
                query2.where = BinaryOperation('and', args=[query2.where, cond])
            else:
                query2.where = cond

        step = self.planner.get_integration_select_step(query2)
        self.tables_fetch_step[item.index] = step

        self.add_plan_step(step)
        self.step_stack.append(step)

    def join_condition_to_columns_map(self, model_table):

        columns_map = {}

        def _check_conditions(node, **kwargs):
            if not isinstance(node, BinaryOperation):
                return

            arg1, arg2 = node.args
            if not (isinstance(arg1, Identifier) and isinstance(arg2, Identifier)):
                return

            table1 = self.get_table_for_column(arg1)
            table2 = self.get_table_for_column(arg2)

            if table1 is model_table:
                # model is on the left
                columns_map[arg1.parts[-1]] = arg2
            elif table2 is model_table:
                # model is on the right
                columns_map[arg2.parts[-1]] = arg1
            else:
                # not found, skip
                return

            # exclude condition
            node.args = [Constant(0), Constant(0)]

        query_traversal(model_table.join_condition, _check_conditions)
        return columns_map

    def get_filters_from_join_conditions(self, fetch_table):

        binary_ops = set()
        conditions = []
        data_conditions = []

        def _check_conditions(node, **kwargs):
            if not isinstance(node, BinaryOperation):
                return

            if node.op != '=':
                binary_ops.add(node.op.lower())
                return

            arg1, arg2 = node.args
            table1 = self.get_table_for_column(arg1) if isinstance(arg1, Identifier) else None
            table2 = self.get_table_for_column(arg2) if isinstance(arg2, Identifier) else None

            if table1 is not fetch_table:
                if table2 is not fetch_table:
                    return
                # set our table first
                table1, table2 = table2, table1
                arg1, arg2 = arg2, arg1

            if isinstance(arg2, Constant):
                conditions.append(node)
            elif table2 is not None:
                data_conditions.append([arg1, arg2])

        query_traversal(fetch_table.join_condition, _check_conditions)

        binary_ops.discard('and')
        if len(binary_ops) > 0:
            # other operations exists, skip
            return []

        for arg1, arg2 in data_conditions:
            # is fetched?
            table2 = self.get_table_for_column(arg2)
            fetch_step = self.tables_fetch_step.get(table2.index)

            if fetch_step is None:
                continue

            # extract distinct values
            # remove aliases
            arg1 = Identifier(parts=[arg1.parts[-1]])
            arg2 = Identifier(parts=[arg2.parts[-1]])

            query2 = Select(targets=[arg2], distinct=True)
            subselect_step = SubSelectStep(query2, fetch_step.result)
            subselect_step = self.add_plan_step(subselect_step)

            conditions.append(BinaryOperation(
                op='in',
                args=[
                    arg1,
                    Parameter(subselect_step.result)
                ]
            ))

        return conditions

    def process_predictor(self, item, query_in):
        if len(self.step_stack) == 0:
            raise NotImplementedError("Predictor can't be first element of join syntax")
        if item.predictor_info.get('timeseries'):
            raise NotImplementedError("TS predictor is not supported here yet")
        data_step = self.step_stack[-1]
        row_dict = None

        predict_target = item.predictor_info.get('to_predict')
        if isinstance(predict_target, list) and len(predict_target) > 0:
            predict_target = predict_target[0]
        if predict_target is not None:
            predict_target = predict_target.lower()

        columns_map = None
        if item.join_condition:
            columns_map = self.join_condition_to_columns_map(item)

        if item.conditions:
            row_dict = {}
            for i, el in enumerate(item.conditions):
                if isinstance(el.args[0], Identifier) and el.op == '=':
                    col_name = el.args[0].parts[-1]
                    if col_name.lower() == predict_target:
                        # don't add predict target to parameters
                        continue

                    if isinstance(el.args[1], (Constant, Parameter)):
                        row_dict[el.args[0].parts[-1]] = el.args[1].value

                    # exclude condition
                    el._orig_node.args = [Constant(0), Constant(0)]

        # params for model
        model_params = None
        partition_size = None
        if query_in.using is not None:
            model_params = {}
            for param, value in query_in.using.items():
                if '.' in param:
                    alias = param.split('.')[0]
                    if (alias,) in item.aliases:
                        new_param = '.'.join(param.split('.')[1:])
                        model_params[new_param.lower()] = value
                else:
                    model_params[param.lower()] = value

            partition_size = model_params.pop('partition_size', None)

        predictor_step = ApplyPredictorStep(
            namespace=item.integration,
            dataframe=data_step.result,
            predictor=item.table,
            params=model_params,
            row_dict=row_dict,
            columns_map=columns_map,
        )

        self.step_stack.append(
            self.add_plan_step(predictor_step, partition_size=partition_size)
        )

    def add_plan_step(self, step, partition_size=None):
        """
        Adds step to plan

        If partition_size is defined: create partition
        If partition is active
            If step can be partitioned:
                Add step to partition not in plan
            Otherwise:
                Add partition to plan
                Add step to plan
        """
        if self.partition:
            if isinstance(step, (JoinStep, ApplyPredictorStep)):
                # add to partition

                self.add_step_to_partition(step)
                return step

        elif partition_size is not None:
            # create partition

            self.partition = MapReduceStep(
                values=step.dataframe,
                reduce='union',
                step=[],
                partition=partition_size
            )
            self.planner.plan.add_step(self.partition)

            self.add_step_to_partition(step)
            return step

        else:
            # next step can't be partitioned.
            self.close_partition()

        return self.planner.plan.add_step(step)

    def add_step_to_partition(self, step):
        step.step_num = f'{self.partition.step_num}_{len(self.partition.step)}'
        self.partition.step.append(step)

    def close_partition(self):
        # return
        # if partitions is exist - clear it and replace last stack item with it

        if self.partition:
            if len(self.step_stack) > 0:
                self.step_stack[-1] = self.partition

            self.partition = None



from typing import Any, Dict, Text

import msal
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb_sql_parser.ast import Constant, Identifier, Select, Star
from mindsdb_sql_parser import parse_sql
import pandas as pd
from requests.exceptions import RequestException

from mindsdb.integrations.handlers.ms_one_drive_handler.ms_graph_api_one_drive_client import MSGraphAPIOneDriveClient
from mindsdb.integrations.handlers.ms_one_drive_handler.ms_one_drive_tables import FileTable, ListFilesTable
from mindsdb.integrations.utilities.handlers.auth_utilities import MSGraphAPIDelegatedPermissionsManager
from mindsdb.integrations.utilities.handlers.auth_utilities.exceptions import AuthException
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    HandlerStatusResponse as StatusResponse,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class MSOneDriveHandler(APIHandler):
    """
    This handler handles the connection and execution of SQL statements on Microsoft OneDrive.
    """

    name = 'one_drive'
    supported_file_formats = ['csv', 'tsv', 'json', 'parquet', 'pdf', 'txt']

    def __init__(self, name: Text, connection_data: Dict, **kwargs: Any) -> None:
        """
        Initializes the handler.

        Args:
            name (Text): The name of the handler instance.
            connection_data (Dict): The connection data required to connect to the Microsoft Graph API.
            kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = connection_data
        self.handler_storage = kwargs['handler_storage']
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def connect(self):
        """
        Establishes a connection to Microsoft OneDrive via the Microsoft Graph API.

        Raises:
            ValueError: If the required connection parameters are not provided.
            AuthenticationError: If an error occurs during the authentication process.

        Returns:
            MSGraphAPIOneDriveClient: An instance of the Microsoft Graph API client for Microsoft OneDrive.
        """
        if self.is_connected and self.connection.check_connection():
            return self.connection

        # Mandatory connection parameters.
        if not all(key in self.connection_data for key in ['client_id', 'client_secret', 'tenant_id']):
            raise ValueError("Required parameters (client_id, client_secret, tenant_id) must be provided.")

        # Initialize the token cache.
        cache = msal.SerializableTokenCache()

        # Load the cache from file if it exists.
        cache_file = 'cache.bin'
        try:
            cache_content = self.handler_storage.file_get(cache_file)
        except FileNotFoundError:
            cache_content = None

        if cache_content:
            cache.deserialize(cache_content)

        # Initialize the Microsoft Authentication Library (MSAL) app.
        permissions_manager = MSGraphAPIDelegatedPermissionsManager(
            client_id=self.connection_data['client_id'],
            client_secret=self.connection_data['client_secret'],
            tenant_id=self.connection_data['tenant_id'],
            cache=cache,
            code=self.connection_data.get('code')
        )

        access_token = permissions_manager.get_access_token()

        # Save the cache back to file if it has changed.
        if cache.has_state_changed:
            self.handler_storage.file_set(cache_file, cache.serialize().encode('utf-8'))

        # Pass the access token to the Microsoft Graph API client for Microsoft OneDrive.
        self.connection = MSGraphAPIOneDriveClient(
            access_token=access_token,
        )

        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the Microsoft Graph API for Microsoft OneDrive.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)

        try:
            connection = self.connect()
            if connection.check_connection():
                response.success = True
                response.copy_storage = True
            else:
                raise RequestException("Connection check failed!")
        except (ValueError, RequestException) as known_error:
            logger.error(f'Connection check to Microsoft OneDrive failed, {known_error}!')
            response.error_message = str(known_error)
        except AuthException as error:
            response.error_message = str(error)
            response.redirect_url = error.auth_url
            return response
        except Exception as unknown_error:
            logger.error(f'Connection check to Microsoft OneDrive failed due to an unknown error, {unknown_error}!')
            response.error_message = str(unknown_error)

        self.is_connected = response.success

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Raises:
            ValueError: If the file format is not supported.
            NotImplementedError: If the query type is not supported.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        if isinstance(query, Select):
            table_name = query.from_table.parts[-1]

            # If the table name is 'files', query the 'files' table.
            if table_name == "files":
                table = ListFilesTable(self)
                df = table.select(query)

            # For any other table name, query the file content via the 'FileTable' class.
            # Only the supported file formats can be queried.
            else:
                extension = table_name.split('.')[-1]
                if extension not in self.supported_file_formats:
                    logger.error(f'The file format {extension} is not supported!')
                    raise ValueError(f'The file format {extension} is not supported!')

                table = FileTable(self, table_name=table_name)
                df = table.select(query)

            return Response(
                RESPONSE_TYPE.TABLE,
                data_frame=df
            )

        else:
            raise NotImplementedError(
                "Only SELECT queries are supported by the Microsoft OneDrive handler."
            )

    def native_query(self, query: Text) -> Response:
        """
        Executes a SQL query and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        query_ast = parse_sql(query)
        return self.query(query_ast)

    def get_tables(self) -> Response:
        """
        Retrieves a list of tables (files) in the user's Microsoft OneDrive.
        Each file is considered a table. Only the supported file formats are included in the list.

        Returns:
            Response: A response object containing the list of tables and views, formatted as per the `Response` class.
        """
        connection = self.connect()

        # Get only the supported file formats.
        # Wrap the file names with backticks to prevent SQL syntax errors.
        supported_files = [
            f"`{file['path']}`"
            for file in connection.get_all_items()
            if file['path'].split('.')[-1] in self.supported_file_formats
        ]

        # Add the 'files' table to the list of supported tables.
        supported_files.insert(0, 'files')

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                supported_files,
                columns=['table_name']
            )
        )

        return response

    def get_columns(self, table_name: str) -> Response:
        """
        Retrieves column details for a specified table (file) in the user's Microsoft OneDrive.

        Args:
            table_name (Text): The name of the table for which to retrieve column information.

        Returns:
            Response: A response object containing the column details, formatted as per the `Response` class.
        """
        # Get the columns (and their data types) by querying a single row from the table.
        query = Select(
            targets=[Star()],
            from_table=Identifier(parts=[table_name]),
            limit=Constant(1)
        )

        result = self.query(query)

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                {
                    'column_name': result.data_frame.columns,
                    'data_type': [data_type if data_type != 'object' else 'string' for data_type in result.data_frame.dtypes]
                }
            )
        )

        return response



import datetime as dt
from typing import Any, Dict, List, Text

from mindsdb_sql_parser.ast import Delete, Insert, Update
import pandas as pd
from slack_sdk.errors import SlackApiError

from mindsdb.integrations.libs.api_handler import APIResource
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions, FilterCondition, FilterOperator, SortColumn
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class SlackConversationsTable(APIResource):
    """
    This is the table abstraction for interacting with conversations via the Slack API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        **kwargs: Any
    ) -> pd.DataFrame:
        """
        Retrieves a list of Slack conversations based on the specified conditions.
        If no channel ID(s) are provided, all channels are retrieved as follows:
        - If the provided limit is greater than 1000, no limit to the API call is provided and the results are paginated until the limit is reached.
        - Otherwise, the provided limit or a default limit of 1000 is used when making the API call.

        Therefore, if a user is to retrieve more than 1000 channels, the limit should be set to a value greater than 1000.

        The above is designed to prevent rate limiting by the Slack API.

        Args:
            conditions (List[FilterCondition]): The conditions to filter the conversations.
            limit (int): The limit of the conversations to return.
            kwargs(Any): Arbitrary keyword arguments.

        Raises:
            ValueError:
                - If an unsupported operator is used for the column 'id'.
                - If the channel ID(s) provided are not found.
            SlackApiError: If an error occurs when getting the channels from the Slack API.

        Returns:
            pd.DataFrame: The list of conversations.
        """
        channels = []
        for condition in conditions:
            value = condition.value
            op = condition.op

            # Handle the column 'id'.
            if condition.column == 'id':
                if op not in [FilterOperator.EQUAL, FilterOperator.IN]:
                    raise ValueError(f"Unsupported operator '{op}' for column 'id'")

                if op == FilterOperator.EQUAL:
                    try:
                        channels = [self.get_channel(value)]
                        condition.applied = True
                    except ValueError:
                        raise

                if op == FilterOperator.IN:
                    try:
                        channels = self._get_channels(
                            value if isinstance(value, list) else [value]
                        )
                        condition.applied = True
                    except ValueError:
                        raise

        # If no channel ID(s) are provided, get all channels with the specified limit.
        if not channels:
            channels = self._get_all_channels(limit)

        for channel in channels:
            channel['created_at'] = dt.datetime.fromtimestamp(channel['created'])
            channel['updated_at'] = dt.datetime.fromtimestamp(channel['updated'] / 1000)

        return pd.DataFrame(channels, columns=self.get_columns())

    def get_channel(self, channel_id: Text) -> Dict:
        """
        Gets the channel data for the specified channel id.

        Args:
            channel_id (Text): The channel id.

        Raises:
            ValueError: If the channel ID is not found.

        Returns:
            Dict: The channel data.
        """
        client = self.handler.connect()

        try:
            response = client.conversations_info(channel=channel_id)
        except SlackApiError as slack_error:
            logger.error(f"Error getting channel '{channel_id}': {slack_error.response['error']}")
            raise ValueError(f"Channel '{channel_id}' not found")

        return response['channel']

    def _get_channels(self, channel_ids: List[Text]) -> List[Dict]:
        """
        Gets the channel data for multiple channel ids.
        As it is unlikely that a large number of channels will be provided, the API rate limits are ignored here.

        Args:
            channel_ids (List[Text]): The list of channel ids.

        Raises:
            ValueError: If a channel ID is not found.

        Returns:
            List[Dict]: The list of channel data.
        """
        channels = []
        for channel_id in channel_ids:
            channel = self.get_channel(channel_id)
            channels.append(channel)

        return channels

    def _get_all_channels(self, limit: int = None) -> List[Dict]:
        """
        Gets the list of channels with a limit.
        If the provided limit is greater than 1000, no limit to the API call is provided and the results are paginated until the limit is reached.
        Otherwise, the provided limit or a default limit of 1000 is used when making the API call.

        Args:
            limit (int): The limit of channels to return.

        Raises:
            SlackApiError: If an error occurs when getting the channels from the Slack API.

        Returns:
            List[Dict]: The list of channels.
        """
        client = self.handler.connect()

        try:
            # If the limit is greater than 1000, paginate the results until the limit is reached.
            if limit and limit > 1000:
                response = client.conversations_list()
                channels = response['channels']

                # Paginate the results until the limit is reached.
                while response['response_metadata']['next_cursor']:
                    response = client.conversations_list(cursor=response['response_metadata']['next_cursor'])
                    channels.extend(response['channels'])
                    if len(channels) >= limit:
                        break

                channels = channels[:limit]
            # Otherwise, use the provided limit or a default limit of 1000.
            else:
                response = client.conversations_list(limit=limit if limit else 1000)
                channels = response['channels']
        except SlackApiError as slack_error:
            logger.error(f"Error getting channels: {slack_error.response['error']}")
            raise

        return channels

    def get_columns(self) -> List[str]:
        """
        Retrieves the attributes (columns) of the Slack conversations.

        Returns:
            List[str]: The list of columns.
        """
        return [
            'id',
            'name',
            'is_channel',
            'is_group',
            'is_im',
            'is_mpim',
            'is_private',
            'is_archived',
            'is_general',
            'is_shared',
            'is_ext_shared',
            'is_org_shared',
            'creator',
            'created_at',
            'updated_at',
        ]


class SlackMessagesTable(APIResource):
    """
    This is the table abstraction for interacting with messages via the Slack API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        **kwargs: Any
    ) -> pd.DataFrame:
        """
        Retrieves a list of messages from Slack conversations based on the specified conditions.

        `channel_id` is a required parameter to retrieve messages from a conversation.

        Messages are retrieved as follows for a given conversation:
        - If the provided limit is greater than 999, no limit to the API call is provided and the results are paginated until the limit is reached.
        - Otherwise, the provided limit or a default limit of 999 is used when making the API call.

        Therefore, if a user is to retrieve more than 999 messages, the limit should be set to a value greater than 999.
        The above is dependent on the other parameters provided in the conditions.

        The above is designed to prevent rate limiting by the Slack API.

        Args:
            conditions (List[FilterCondition]): The conditions to filter the messages.
            limit (int): The limit of the messages to return.
            sort (List[SortColumn]): The columns to sort the messages by.
            kwargs (Any): Arbitrary keyword arguments.

        Raises:
            ValueError:
                - If the 'channel_id' parameter is not provided.
                - If an unsupported operator is used for the column 'channel_id'.
                - If the channel ID provided is not found.
            SlackApiError: If an error occurs when getting the messages from the Slack API.

        Returns:
            pd.DataFrame: The list of messages.
        """
        client = self.handler.connect()

        # Build the parameters for the call to the Slack API.
        params = {}
        for condition in conditions:
            value = condition.value
            op = condition.op

            # Handle the column 'channel_id'.
            if condition.column == 'channel_id':
                if op != FilterOperator.EQUAL:
                    raise ValueError(f"Unsupported operator '{op}' for column 'channel_id'")

                # Check if the provided channel exists.
                try:
                    channel = SlackConversationsTable(self.handler).get_channel(value)
                    params['channel'] = value
                    condition.applied = True
                except SlackApiError:
                    raise ValueError(f"Channel '{value}' not found")

            # Handle the column 'created_at'.
            elif condition.column == 'created_at' and value is not None:
                date = dt.datetime.fromisoformat(value).replace(tzinfo=dt.timezone.utc)
                if op == FilterOperator.GREATER_THAN:
                    params['oldest'] = date.timestamp() + 1
                elif op == FilterOperator.GREATER_THAN_OR_EQUAL:
                    params['oldest'] = date.timestamp()
                elif op == FilterOperator.LESS_THAN_OR_EQUAL:
                    params['latest'] = date.timestamp()
                else:
                    continue
                condition.applied = True

        if 'channel' not in params:
            raise ValueError("To retrieve data from Slack, you need to provide the 'channel_id' parameter.")

        # Retrieve the messages from the Slack API.
        try:
            # If the limit is greater than 999, paginate the results until the limit is reached.
            if limit and limit > 999:
                response = client.conversations_history(**params)
                messages = response['messages']

                # Paginate the results until the limit is reached.
                while response['response_metadata']['next_cursor']:
                    response = client.conversations_history(cursor=response['response_metadata']['next_cursor'])
                    messages.extend(response['messages'])
                    if len(messages) >= limit:
                        break

                    messages = messages[:limit]
            # Otherwise, use the provided limit or a default limit of 999.
            else:
                params['limit'] = limit if limit else 999
                response = client.conversations_history(**params)
                messages = response['messages']
        except SlackApiError as slack_error:
            logger.error(f"Error getting messages: {slack_error.response['error']}")
            raise

        result = pd.DataFrame(messages, columns=self.get_columns())

        result = result[result['text'].notnull()]

        # Add the channel ID and name to the result.
        result['channel_id'] = params['channel']
        result['channel_name'] = channel['name'] if 'name' in channel else None

        # Translate the time stamp into a 'created_at' field.
        result['created_at'] = pd.to_datetime(result['ts'].astype(float), unit='s').dt.strftime('%Y-%m-%d %H:%M:%S')

        # Sort the messages by the specified columns.
        if sort:
            result.sort_values(
                by=[col.column for col in sort],
                ascending=[col.ascending for col in sort],
                inplace=True
            )

        return result

    def insert(self, query: Insert):
        """
        Executes an INSERT SQL query represented by an ASTNode object and posts a message to a Slack channel.

        Args:
            query (Insert): An ASTNode object representing the SQL query to be executed.

        Raises:
            ValueError: If the 'channel_id' or 'text' parameters are not provided.
            SlackApiError: If an error occurs when posting the message to the Slack channel.
        """
        client = self.handler.connect()

        # Get column names and values from the query.
        columns = [col.name for col in query.columns]
        for row in query.values:
            params = dict(zip(columns, row))

            # Check if required parameters are provided.
            if 'channel_id' not in params or 'text' not in params:
                raise ValueError("To insert data into Slack, you need to provide the 'channel_id' and 'text' parameters.")

            try:
                client.chat_postMessage(
                    channel=params['channel_id'],
                    text=params['text']
                )
            except SlackApiError as slack_error:
                logger.error(f"Error posting message to Slack channel '{params['channel_id']}': {slack_error.response['error']}")
                raise

    def update(self, query: Update):
        """
        Executes an UPDATE SQL query represented by an ASTNode object and updates a message in a Slack channel.

        Args:
            query (Update): An ASTNode object representing the SQL query to be executed.

        Raises:
            ValueError:
                - If the 'channel_id', 'ts', or 'text' parameters are not provided.
                - If an unsupported operator is used for the columns.
                - If an unsupported column is used.
                - If the channel ID provided is not found.
            SlackApiError: If an error occurs when updating the message in the Slack channel.
        """
        client = self.handler.connect()

        conditions = extract_comparison_conditions(query.where)

        # Build the parameters for the call to the Slack API.
        params = {}
        # Extract the parameters from the conditions.
        for op, arg1, arg2 in conditions:
            # Handle the column 'channel_id'.
            if arg1 == 'channel_id':
                # Check if the provided channel exists.
                try:
                    SlackConversationsTable(self.handler).get_channel(arg2)
                    params['channel'] = arg2
                except SlackApiError as slack_error:
                    logger.error(f"Error getting channel '{arg2}': {slack_error.response['error']}")
                    raise ValueError(f"Channel '{arg2}' not found")

            # Handle the column'ts'.
            elif arg1 == 'ts':
                if op == '=':
                    params[arg1] = str(arg2)
                else:
                    raise ValueError(f"Unsupported operator '{op}' for column '{arg1}'")

            else:
                raise ValueError(f"Unsupported column '{arg1}'")

        # Extract the update columns and values.
        for col, val in query.update_columns.items():
            if col == 'text':
                params[col] = str(val).strip("'")
            else:
                raise ValueError(f"Unsupported column '{col}'")

        # Check if required parameters are provided.
        if 'channel' not in params or 'ts' not in params or 'text' not in params:
            raise ValueError("To update a message in Slack, you need to provide the 'channel', 'ts', and 'text' parameters.")

        try:
            client.chat_update(
                channel=params['channel'],
                ts=params['ts'],
                text=params['text'].strip()
            )
        except SlackApiError as slack_error:
            logger.error(f"Error updating message in Slack channel '{params['channel']}' with timestamp '{params['ts']}': {slack_error.response['error']}")
            raise

    def delete(self, query: Delete):
        """
        Executes a DELETE SQL query represented by an ASTNode object and deletes a message from a Slack channel.

        Args:
            query (Delete): An ASTNode object representing the SQL query to be executed.

        Raises:
            ValueError:
                - If the 'channel_id' or 'ts' parameters are not provided.
                - If an unsupported operator is used for the columns.
                - If an unsupported column is used.
                - If the channel ID provided is not found.
            SlackApiError: If an error occurs when deleting the message from the Slack channel.
        """
        client = self.handler.connect()

        conditions = extract_comparison_conditions(query.where)

        # Build the parameters for the call to the Slack API.
        params = {}
        for op, arg1, arg2 in conditions:
            # Handle the column 'channel_id'.
            if arg1 == 'channel_id':
                # Check if the provided channel exists.
                try:
                    SlackConversationsTable(self.handler).get_channel(arg2)
                    params['channel'] = arg2
                except SlackApiError as slack_error:
                    logger.error(f"Error getting channel '{arg2}': {slack_error.response['error']}")
                    raise ValueError(f"Channel '{arg2}' not found")

            # Handle the columns 'ts'.
            elif arg1 == 'ts':
                if op == '=':
                    params['ts'] = float(arg2)
                else:
                    raise NotImplementedError(f'Unknown op: {op}')

            else:
                raise ValueError(f"Unsupported column '{arg1}'")

        # Check if required parameters are provided.
        if 'channel' not in params or 'ts' not in params:
            raise ValueError("To delete a message from Slack, you need to provide the 'channel' and 'ts' parameters.")

        try:
            client.chat_delete(
                channel=params['channel'],
                ts=params['ts']
            )

        except SlackApiError as slack_error:
            logger.error(f"Error deleting message in Slack channel '{params['channel']}' with timestamp '{params['ts']}': {slack_error.response['error']}")
            raise

    def get_columns(self) -> List[Text]:
        """
        Retrieves the attributes (columns) of the Slack messages.

        Returns:
            List[str]: The list of columns.
        """
        return [
            'channel_id',
            'channel',
            'client_msg_id',
            'type',
            'subtype',
            'ts',
            'created_at',
            'user',
            'text',
            'attachments',
            'files',
            'reactions',
            'thread_ts',
            'reply_count',
            'reply_users_count',
            'latest_reply',
            'reply_users'
        ]


class SlackThreadsTable(APIResource):
    """
    This is the table abstraction for interacting with threads in Slack conversations.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        **kwargs: Any
    ) -> pd.DataFrame:
        """
        Retrieves a list of messages in a thread based on the specified conditions.

        `channel_id` and `thread_ts` are required parameters to retrieve messages from a thread.

        Messages are retrieved as follows for a given thread:
        - If the provided limit is greater than 1000, no limit to the API call is provided and the results are paginated until the limit is reached.
        - Otherwise, the provided limit or a default limit of 1000 is used when making the API call.

        Therefore, if a user is to retrieve more than 1000 messages, the limit should be set to a value greater than 1000.

        The above is designed to prevent rate limiting by the Slack API.

        Args:
            conditions (List[FilterCondition]): The conditions to filter the messages.
            limit (int): The limit of the messages to return.
            sort (List[SortColumn]): The columns to sort the messages by.
            kwargs (Any): Arbitrary keyword arguments.

        Raises:
            ValueError:
                - If the 'channel_id' or 'thread_ts' parameters are not provided.
                - If an unsupported operator is used for the columns.
                - If an unsupported column is used.
                - If the channel ID provided is not found.
            SlackApiError: If an error occurs when getting the messages from the Slack API.

        Returns:
            pd.DataFrame: The messages in the thread.
        """
        client = self.handler.connect()

        # Build the parameters for the call to the Slack API.
        params = {}
        for condition in conditions:
            value = condition.value
            op = condition.op

            # Handle the column 'channel_id'.
            if condition.column == 'channel_id':
                if op != FilterOperator.EQUAL:
                    raise ValueError(f"Unsupported operator '{op}' for column 'channel_id'")

                # Check if the provided channel exists.
                try:
                    channel = SlackConversationsTable(self.handler).get_channel(value)
                    params['channel'] = value
                    condition.applied = True
                except SlackApiError as slack_error:
                    logger.error(f"Error getting channel '{value}': {slack_error.response['error']}")
                    raise ValueError(f"Channel '{value}' not found")

            # Handle the column 'thread_ts'.
            elif condition.column == 'thread_ts':
                if op != FilterOperator.EQUAL:
                    raise ValueError(f"Unsupported operator '{op}' for column 'thread_ts'")

                params['ts'] = value

        if 'channel' not in params or 'ts' not in params:
            raise ValueError("To retrieve data from Slack, you need to provide the 'channel_id' and 'thread_ts' parameters.")

        # Retrieve the messages from the Slack API.
        try:
            # If the limit is greater than 1000, paginate the results until the limit is reached.
            if limit and limit > 1000:
                response = client.conversations_replies(**params)
                messages = response['messages']

                # Paginate the results until the limit is reached.
                while response['response_metadata']['next_cursor']:
                    response = client.conversations_replies(cursor=response['response_metadata']['next_cursor'])
                    messages.extend(response['messages'])
                    if len(messages) >= limit:
                        break

                    messages = messages[:limit]
            # Otherwise, use the provided limit or a default limit of 1000.
            else:
                params['limit'] = limit if limit else 1000
                response = client.conversations_replies(**params)
                messages = response['messages']
        except SlackApiError as slack_error:
            logger.error(f"Error getting messages: {slack_error.response['error']}")
            raise

        result = pd.DataFrame(messages, columns=self.get_columns())

        result = result[result['text'].notnull()]

        # Add the channel ID and name to the result.
        result['channel_id'] = params['channel']
        result['channel_name'] = channel['name'] if 'name' in channel else None

        # Sort the messages by the specified columns.
        if sort:
            result.sort_values(
                by=[col.column for col in sort],
                ascending=[col.ascending for col in sort],
                inplace=True
            )

        return result

    def insert(self, query: Insert):
        """
        Executes an INSERT SQL query represented by an ASTNode object and posts a message to a Slack thread.

        Args:
            query (Insert): An ASTNode object representing the SQL query to be executed.

        Raises:
            ValueError: If the 'channel_id', 'text', or 'thread_ts' parameters are not provided.
        """
        client = self.handler.connect()

        # Get column names and values from the query.
        columns = [col.name for col in query.columns]
        for row in query.values:
            params = dict(zip(columns, row))

            # Check if required parameters are provided.
            if 'channel_id' not in params or 'text' not in params or 'thread_ts' not in params:
                raise ValueError("To insert data into Slack, you need to provide the 'channel_id', 'text', and 'thread_ts' parameters.")

            try:
                client.chat_postMessage(
                    channel=params['channel_id'],
                    text=params['text'],
                    thread_ts=params['thread_ts']
                )
            except SlackApiError as slack_error:
                logger.error(f"Error posting message to Slack channel '{params['channel_id']}': {slack_error.response['error']}")
                raise

    def get_columns(self) -> List[Text]:
        """
        Retrieves the attributes (columns) of the Slack threads.

        Returns:
            List[Text]: The list of columns.
        """
        return [
            "channel_id",
            "channel_name",
            "type",
            "user",
            "text",
            "ts",
            "client_msg_id",
            "thread_ts",
            "parent_user_id",
            "reply_count",
            "reply_users_count",
            "latest_reply",
            "reply_users"
        ]


class SlackUsersTable(APIResource):
    """
    This is the table abstraction for interacting with users in Slack.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        **kwargs: Any
    ) -> pd.DataFrame:
        """
        Retrieves a list of users based on the specified conditions.
        Users are retrieved as follows:
        - If the provided limit is greater than 1000, no limit to the API call is provided and the results are paginated until the limit is reached.
        - Otherwise, the provided limit or a default limit of 1000 is used when making the API call.

        Therefore, if a user is to retrieve more than 1000 users, the limit should be set to a value greater than 1000.

        The above is designed to prevent rate limiting by the Slack API.

        Args:
            conditions (List[FilterCondition]): The conditions to filter the users.
            limit (int): The limit of the users to return.
            kwargs (Any): Arbitrary keyword arguments.

        Raises:
            SlackApiError: If an error occurs when getting the users from the Slack API.
        """
        client = self.handler.connect()

        # Retrieve the users from the Slack API.
        try:
            # If the limit is greater than 1000, paginate the results until the limit is reached.
            if limit and limit > 1000:
                response = client.users_list()
                users = response['members']

                # Paginate the results until the limit is reached.
                while response['response_metadata']['next_cursor']:
                    response = client.users_list(cursor=response['response_metadata']['next_cursor'])
                    users.extend(response['members'])
                    if len(users) >= limit:
                        break

                    users = users[:limit]
            # Otherwise, use the provided limit or a default limit of 1000.
            else:
                response = client.users_list(limit=limit if limit else 1000)
                users = response['members']
        except SlackApiError as slack_error:
            logger.error(f"Error getting users: {slack_error.response['error']}")
            raise

        return pd.DataFrame(users, columns=self.get_columns())

    def get_columns(self) -> List[Text]:
        """
        Retrieves the attributes (columns) of the Slack users.

        Returns:
            List[Text]: The list of columns.
        """
        return [
            'id',
            'name',
            'real_name'
        ]



from collections import defaultdict

from mindsdb_sql_parser.ast import (
    Identifier,
    Select,
    Star,
)
from mindsdb.api.executor.planner.steps import ProjectStep
from mindsdb.integrations.utilities.query_traversal import query_traversal

from mindsdb.api.executor.sql_query.result_set import ResultSet
from mindsdb.api.executor.utilities.sql import query_df
from mindsdb.api.executor.exceptions import (
    KeyColumnDoesNotExist,
    NotSupportedYet
)

from .base import BaseStepCall


class ProjectStepCall(BaseStepCall):

    bind = ProjectStep

    def call(self, step):
        result_set = self.steps_data[step.dataframe.step_num]

        df, col_names = result_set.to_df_cols()
        col_idx = {}
        tbl_idx = defaultdict(list)
        for name, col in col_names.items():
            col_idx[col.alias] = name
            col_idx[(col.table_alias, col.alias)] = name
            # add to tables
            tbl_idx[col.table_name].append(name)
            if col.table_name != col.table_alias:
                tbl_idx[col.table_alias].append(name)

        # analyze condition and change name of columns
        def check_fields(node, is_table=None, **kwargs):
            if is_table:
                raise NotSupportedYet('Subqueries is not supported in target')
            if isinstance(node, Identifier):
                # only column name
                col_name = node.parts[-1]
                if isinstance(col_name, Star):
                    if len(node.parts) == 1:
                        # left as is
                        return
                    else:
                        # replace with all columns from table
                        table_name = node.parts[-2]
                        return [
                            Identifier(parts=[col])
                            for col in tbl_idx.get(table_name, [])
                        ]

                if len(node.parts) == 1:
                    key = col_name
                else:
                    table_name = node.parts[-2]
                    key = (table_name, col_name)

                if key not in col_idx:
                    raise KeyColumnDoesNotExist(f'Table not found for column: {key}')

                new_name = col_idx[key]
                return Identifier(parts=[new_name], alias=node.alias)

        query = Select(
            targets=step.columns,
            from_table=Identifier('df_table')
        )

        targets0 = query_traversal(query.targets, check_fields)
        targets = []
        for target in targets0:
            if isinstance(target, list):
                targets.extend(target)
            else:
                targets.append(target)
        query.targets = targets

        res = query_df(df, query, session=self.session)

        return ResultSet().from_df_cols(res, col_names, strict=False)



import pandas as pd
from typing import Text, List, Dict, Any

from mindsdb_sql_parser import ast
from mindsdb.integrations.libs.api_handler import APITable

from mindsdb.integrations.utilities.handlers.query_utilities.select_query_utilities import (
    SELECTQueryParser,
    SELECTQueryExecutor,
)
from mindsdb.integrations.utilities.handlers.query_utilities.delete_query_utilities import (
    DELETEQueryParser,
    DELETEQueryExecutor,
)

from mindsdb.integrations.utilities.handlers.query_utilities.update_query_utilities import (
    UPDATEQueryParser,
    UPDATEQueryExecutor,
)

from mindsdb.integrations.utilities.handlers.query_utilities import INSERTQueryParser


class SitesTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Sharepoint Sites data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Sharepoint Sites matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(query, "sites", self.get_columns())
        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        sites_df = pd.json_normalize(self.get_sites(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            sites_df, selected_columns, where_conditions, order_by_conditions
        )
        sites_df = select_statement_executor.execute_query()

        return sites_df

    def update(self, query: ast.Update) -> None:
        """Updates data in the Sharepoint "PUT /lists" API endpoint.

        Parameters
        ----------
        query : ast.Update
           Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()

        sites_df = pd.json_normalize(self.get_sites())

        update_query_executor = UPDATEQueryExecutor(sites_df, where_conditions)

        sites_df = update_query_executor.execute_query()

        sites_ids = sites_df[["siteId"]].to_dict(orient="records")

        self.update_sites(sites_ids, values_to_update)

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_sites(limit=1)).columns.tolist()

    def get_sites(self, **kwargs) -> List[Dict]:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        site_data = client.get_all_sites(**kwargs)
        return site_data

    def update_sites(self, site_ids: List[dict], values_to_update: dict) -> None:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        client.update_sites(site_ids, values_to_update)


class ListsTable(APITable):
    def insert(self, query: ast.Insert) -> None:
        """Inserts data into the Sharepoint "POST /lists" API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=["displayName", "columns", "list", "siteId"],
            mandatory_columns=["displayName", "list", "siteId"],
            all_mandatory=False,
        )
        lists_data = insert_statement_parser.parse_query()
        self.create_lists(lists_data)

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Sharepoint lists data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Sharepoint lists matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(query, "lists", self.get_columns())
        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        lists_df = pd.json_normalize(self.get_lists(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            lists_df, selected_columns, where_conditions, order_by_conditions
        )
        lists_df = select_statement_executor.execute_query()

        return lists_df

    def update(self, query: ast.Update) -> None:
        """Updates data in the Sharepoint "PUT /lists" API endpoint.

        Parameters
        ----------
        query : ast.Update
           Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()

        lists_df = pd.json_normalize(self.get_lists())

        update_query_executor = UPDATEQueryExecutor(lists_df, where_conditions)

        lists_df = update_query_executor.execute_query()

        list_ids = lists_df[["id", "siteId"]].to_dict(orient="records")

        self.update_lists(list_ids, values_to_update)

    def delete(self, query: ast.Delete) -> None:
        """
        Deletes data from the Sharepoint "DELETE /lists" API endpoint.

        Parameters
        ----------
        query : ast.Delete
           Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()

        lists_df = pd.json_normalize(self.get_lists())

        delete_query_executor = DELETEQueryExecutor(lists_df, where_conditions)

        lists_df = delete_query_executor.execute_query()

        list_ids = lists_df[["id", "siteId"]].to_dict(orient="records")
        self.delete_lists(list_ids)

    def get_lists(self, **kwargs) -> List[Dict]:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        lists_data = client.get_all_lists(**kwargs)
        return lists_data

    def delete_lists(self, list_ids: List[dict]) -> None:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        client.delete_lists(list_ids)

    def update_lists(self, list_ids: List[dict], values_to_update: dict) -> None:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        client.update_lists(list_ids, values_to_update)

    def create_lists(self, lists_data: List[Dict[Text, Any]]) -> None:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        client.create_lists(data=lists_data)

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_lists(limit=1)).columns.tolist()


class SiteColumnsTable(APITable):
    def insert(self, query: ast.Insert) -> None:
        """Inserts data into the Sharepoint "POST /columns" API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=[
                "text",
                "name",
                "indexed",
                "enforceUniqueValues",
                "hidden",
                "siteId",
            ],
            mandatory_columns=["name", "siteId"],
            all_mandatory=False,
        )
        site_columns_data = insert_statement_parser.parse_query()
        self.create_site_columns(site_columns_data)

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Sharepoint columns data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Sharepoint Columns matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(
            query, "siteColumns", self.get_columns()
        )
        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        site_columns_df = pd.json_normalize(self.get_site_columns(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            site_columns_df,
            selected_columns,
            where_conditions,
            order_by_conditions,
        )
        site_columns_df = select_statement_executor.execute_query()

        return site_columns_df

    def update(self, query: ast.Update) -> None:
        """Updates data in the Sharepoint "PUT /columns" API endpoint.

        Parameters
        ----------
        query : ast.Update
           Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()

        site_columns_df = pd.json_normalize(self.get_site_columns())

        update_query_executor = UPDATEQueryExecutor(site_columns_df, where_conditions)

        site_columns_df = update_query_executor.execute_query()

        site_columns_ids = site_columns_df[["id", "siteId"]].to_dict(orient="records")

        self.update_site_columns(site_columns_ids, values_to_update)

    def delete(self, query: ast.Delete) -> None:
        """
        Deletes data from the Sharepoint "DELETE /columns" API endpoint.

        Parameters
        ----------
        query : ast.Delete
           Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()

        site_columns_df = pd.json_normalize(self.get_site_columns())

        delete_query_executor = DELETEQueryExecutor(site_columns_df, where_conditions)

        site_columns_df = delete_query_executor.execute_query()

        site_columns_ids = site_columns_df[["id", "siteId"]].to_dict(orient="records")
        self.delete_site_columns(site_columns_ids)

    def get_site_columns(self, **kwargs) -> List[Dict]:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        site_columns_data = client.get_all_site_columns(**kwargs)
        return site_columns_data

    def delete_site_columns(self, sharepoint_column_ids: List[dict]) -> None:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        client.delete_site_columns(sharepoint_column_ids)

    def update_site_columns(
        self, sharepoint_column_ids: List[dict], values_to_update: dict
    ) -> None:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        client.update_site_columns(sharepoint_column_ids, values_to_update)

    def create_site_columns(
        self, sharepoint_column_data: List[Dict[Text, Any]]
    ) -> None:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        client.create_site_columns(data=sharepoint_column_data)

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_site_columns(limit=1)).columns.tolist()


class ListItemsTable(APITable):
    def insert(self, query: ast.Insert) -> None:
        """Inserts data into the Sharepoint "POST /items" API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=["fields", "listId", "siteId"],
            mandatory_columns=["listId", "siteId"],
            all_mandatory=False,
        )
        list_items_data = insert_statement_parser.parse_query()
        self.create_list_items(list_items_data)

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Sharepoint items data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Sharepoint items matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(
            query, "listItems", self.get_columns()
        )
        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        list_items_df = pd.json_normalize(self.get_list_items(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            list_items_df, selected_columns, where_conditions, order_by_conditions
        )
        list_items_df = select_statement_executor.execute_query()

        return list_items_df

    def update(self, query: ast.Update) -> None:
        """Updates data in the Sharepoint "PUT /items" API endpoint.

        Parameters
        ----------
        query : ast.Update
           Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()

        list_items_df = pd.json_normalize(self.get_list_items())

        update_query_executor = UPDATEQueryExecutor(list_items_df, where_conditions)

        list_items_df = update_query_executor.execute_query()

        list_items_ids = list_items_df[["id", "siteId", "listId"]].to_dict(
            orient="records"
        )

        self.update_list_items(list_items_ids, values_to_update)

    def delete(self, query: ast.Delete) -> None:
        """
        Deletes data from the Sharepoint "DELETE /items" API endpoint.

        Parameters
        ----------
        query : ast.Delete
           Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()

        list_items_df = pd.json_normalize(self.get_list_items())

        delete_query_executor = DELETEQueryExecutor(list_items_df, where_conditions)

        list_items_df = delete_query_executor.execute_query()

        list_items_ids = list_items_df[["id", "siteId", "listId"]].to_dict(
            orient="records"
        )
        self.delete_list_items(list_items_ids)

    def get_list_items(self, **kwargs) -> List[Dict]:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        list_items_data = client.get_all_items(**kwargs)
        return list_items_data

    def delete_list_items(self, list_item_ids: List[dict]) -> None:
        if not self.handler.connnection.check_connection():
            self.handler.connect()
        client = self.handler.connection
        client.delete_items(list_item_ids)

    def update_list_items(
        self, list_items_ids: List[dict], values_to_update: dict
    ) -> None:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        client.update_items(item_dict=list_items_ids, values_to_update=values_to_update)

    def create_list_items(self, list_items_data: List[Dict[Text, Any]]) -> None:
        if not self.handler.connection.check_bearer_token_validity():
            self.handler.connect()
        client = self.handler.connection
        client.create_items(data=list_items_data)

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_list_items(limit=1)).columns.tolist()



"""added webhook_token to chat_bots

Revision ID: 6c57ed39a82b
Revises: 8e17ff6b75e9
Create Date: 2024-10-07 16:40:14.141878

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = '6c57ed39a82b'
down_revision = '8e17ff6b75e9'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('chat_bots', schema=None) as batch_op:
        batch_op.add_column(sa.Column('webhook_token', sa.VARCHAR(), nullable=True))


def downgrade():
    with op.batch_alter_table('chat_bots', schema=None) as batch_op:
        batch_op.drop_column('webhook_token')



"""llm_log_json_in_out

Revision ID: a8a3fac369e7
Revises: 0f89b523f346
Create Date: 2024-11-28 17:19:20.798803

"""
import json

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import table
import mindsdb.interfaces.storage.db  # noqa

# revision identifiers, used by Alembic.
revision = 'a8a3fac369e7'
down_revision = '0f89b523f346'
branch_labels = None
depends_on = None


def upgrade():
    llm_log_table = table(
        'llm_log',
        sa.Column('id', sa.Integer),
        sa.Column('input', sa.String),
        sa.Column('output', sa.String),
        sa.Column('input_json', sa.JSON),
        sa.Column('output_json', sa.JSON)
    )

    with op.batch_alter_table('llm_log', schema=None) as batch_op:
        batch_op.add_column(sa.Column('input_json', sa.JSON(), nullable=True))
        batch_op.add_column(sa.Column('output_json', sa.JSON(), nullable=True))

    connection = op.get_bind()
    for row in connection.execute(llm_log_table.select()):
        try:
            input_json = json.loads(row.input)
        except Exception:
            input_json = None

        output_json = None
        try:
            if row.output is not None:
                output_json = [str(row.output)]
        except Exception:
            pass

        connection.execute(
            llm_log_table.update().where(
                llm_log_table.c.id == row.id
            ).values(input_json=input_json, output_json=output_json)
        )

    with op.batch_alter_table('llm_log', schema=None) as batch_op:
        batch_op.drop_column('input')
        batch_op.alter_column('input_json', new_column_name='input')
        batch_op.drop_column('output')
        batch_op.alter_column('output_json', new_column_name='output')


def downgrade():
    llm_log_table = table(
        'llm_log',
        sa.Column('id', sa.Integer),
        sa.Column('input', sa.JSON),
        sa.Column('output', sa.JSON),
        sa.Column('input_str', sa.String),
        sa.Column('output_str', sa.String)
    )

    with op.batch_alter_table('llm_log', schema=None) as batch_op:
        batch_op.add_column(sa.Column('input_str', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('output_str', sa.String(), nullable=True))

    connection = op.get_bind()
    for row in connection.execute(llm_log_table.select()):
        input_str = None
        if row.input is not None:
            try:
                input_str = json.dumps(row.input)
            except Exception:
                pass

        output_str = None
        if isinstance(row.output, list):
            try:
                output_str = '\n'.join(row.output)
            except Exception:
                pass

        connection.execute(
            llm_log_table.update().where(
                llm_log_table.c.id == row.id
            ).values(input_str=input_str, output_str=output_str)
        )

    with op.batch_alter_table('llm_log', schema=None) as batch_op:
        batch_op.drop_column('input')
        batch_op.alter_column('input_str', new_column_name='input')
        batch_op.drop_column('output')
        batch_op.alter_column('output_str', new_column_name='output')



from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from collections import OrderedDict


from mindsdb.integrations.handlers.mysql_handler import Handler as MysqlHandler


class VitessHandler(MysqlHandler):
    """
    This handler handles connection and execution of the Vitess statements.
    """
    name = 'vitess'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)



connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Vitess server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the Vitess server.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the Vitess server.'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Vitess server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the Vitess server. Must be an integer.'
    }
)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=5432,
    user='root',
    password='<???>',
    database='database'
)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The Snowflake host.',
        'required': True,
        'label': 'Host'
    },
    account={
        'type': ARG_TYPE.STR,
        'description': 'The Snowflake account name.',
        'required': False,
        'label': 'Server'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Snowflake account.',
        'required': True,
        'label': 'User'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Snowflake account.',
        'required': True,
        'label': 'Password',
        'secret': True
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database to use when connecting to the Snowflake account.',
        'required': True,
        'label': 'Database'
    },
    schema={
        'type': ARG_TYPE.STR,
        'description': 'The schema to use when connecting to the Snowflake account.',
        'required': False,
        'label': 'Schema'
    },
    warehouse={
        'type': ARG_TYPE.STR,
        'description': 'The warehouse to use when executing queries on the Snowflake account.',
        'required': False,
        'label': 'Warehouse'
    },
    role={
        'type': ARG_TYPE.STR,
        'description': 'The role to use when executing queries on the Snowflake account.',
        'required': False,
        'label': 'Role'
    }
)

connection_args_example = OrderedDict(
    account='abcxyz-1234567',
    user='user',
    password='password',
    database='test'
)



from typing import List
from textwrap import dedent

from langchain_community.agent_toolkits.sql.toolkit import SQLDatabaseToolkit
from langchain_community.tools import ListSQLDatabaseTool, InfoSQLDatabaseTool, QuerySQLDataBaseTool
from langchain_core.tools import BaseTool

from mindsdb.interfaces.skills.custom.text2sql.mindsdb_sql_tool import MindsDBSQLParserTool


class MindsDBSQLToolkit(SQLDatabaseToolkit):

    def get_tools(self, prefix='') -> List[BaseTool]:
        """Get the tools in the toolkit."""
        list_sql_database_tool = ListSQLDatabaseTool(
            name=f'sql_db_list_tables{prefix}',
            db=self.db,
            description=dedent("""\n
                Input is an empty string, output is a comma-separated list of tables in the database. Each table name is escaped using backticks.
                Each table name in the list may be in one of two formats: database_name.`table_name` or database_name.schema_name.`table_name`.
                Table names in response to the user must be escaped using backticks.
            """)
        )

        info_sql_database_tool_description = (
            "Input: A comma-separated list of tables enclosed between the symbols $START$ and $STOP$. The tables names itself must be escaped using backticks.\n"
            "Output: Schema and sample rows for those tables. \n"
            "Use this tool to investigate table schemas for needed columns. "
            f"Ensure tables exist by calling {list_sql_database_tool.name} first. "
            # "The names of tables, schemas, and databases must be escaped using backticks. "
            # "Always enclose the names of tables, schemas, and databases in backticks. "
            "Get sample data with 'SELECT * FROM `database`.`table` LIMIT 3' before answering questions. \n"
            "Example of correct Input:\n    $START$ `database`.`table1`, `database`.`table2`, `database`.`table3` $STOP$\n"
            "    $START$ `table1` `table2` `table3` $STOP$\n"
            "Example of wrong Input:\n    $START$ `database.table1`, `database.table2`, `database.table3` $STOP$\n"
            "    $START$ table1 table2 table3 $STOP$\n"
        )
        info_sql_database_tool = InfoSQLDatabaseTool(
            name=f'sql_db_schema{prefix}',
            db=self.db, description=info_sql_database_tool_description
        )

        query_sql_database_tool_description = dedent(f"""\
            Input: A detailed and well-structured SQL query. The query must be enclosed between the symbols $START$ and $STOP$.
            Output: Database result or error message. For errors, rewrite and retry the query. For 'Unknown column' errors, use '{info_sql_database_tool.name}' to check table fields.
            This system is a highly intelligent and reliable PostgreSQL SQL skill designed to work with databases.
            Follow these instructions with utmost precision:
            1. Query Output Format:
               - Always return results in well-formatted **Markdown tables**.
               - Ensure clarity and proper structure for easy readability.
            2. Sample Data:
               - Before answering a question, if you don't have sample data about a table, **always** get sample data using `SELECT * FROM table LIMIT 3` from the tables you believe are relevant to formulating your answers.
            3. Categorical Data:
               - Whenever working with a column where values seem categorical, especially when filtering with `WHERE col = 'value'`, `WHERE col IN (list of values)`, or `WHERE col NOT IN (list of values)`, **always** retrieve the distinct values first.
               - Before writing your main query, always run `SELECT DISTINCT col` to fetch a list of unique values from that column. This step is mandatory to ensure accurate queries and responses.
            4. Result Limiting and Counting:
               - Unless instructed otherwise by the user, always run a count on the final query first using `SELECT COUNT(*)`.
               - If the count is greater than 10, limit the query to return only 10 results initially.
               - **Always** inform the user of the total number of results available and specify that you are providing the first 10 results.
               - Let the user know they can request additional results and/or specify how they would like the results ordered or grouped.
            5. Date Handling:
               - **Always** use PostgreSQL-compatible `CURRENT_DATE` or `NOW()` functions when working with dates—never assume or guess the current date.
               - For any date-related comparisons in the query, *always* ensure that your query casts the column being compared using `column_name::DATE [operator] ..`
               - Do not compare date values without casting columns to date.
               - For date interval operations, use Interval units as keywords. You can use keywords to specify units like days, hours, months, years, etc., directly without quotes. Examples:
                 SELECT NOW() + INTERVAL 5 DAY;
                 SELECT NOW() - INTERVAL 3 HOUR;
                 SELECT NOW() + INTERVAL 2 MONTH + INTERVAL 3 DAY;
                 SELECT NOW() - INTERVAL 1 YEAR;
            6. Query Best Practices:
               - Always send only one query at a time.
               - Always enclose the names of tables, schemas, and databases in backticks.
               - The input SQL query must end with a semicolon.
               - Query only necessary columns, not all.
               - Use only existing column names from correct tables.
               - Use database-specific syntax for date operations.
            7. Error Handling:
               - For errors, rewrite and retry the query.
               - For 'Unknown column' errors, check table fields using info_sql_database_tool.
            Adhere to these guidelines for all queries and responses. Ask for clarification if needed.
        """)

        query_sql_database_tool = QuerySQLDataBaseTool(
            name=f'sql_db_query{prefix}',
            db=self.db, description=query_sql_database_tool_description
        )

        mindsdb_sql_parser_tool_description = (
            "Use this tool to ensure a SQL query passes the MindsDB SQL parser. "
            "If the query is not correct, it will be corrected and returned. Use the new query. "
            "If the query can't be corrected, an error is returned. In this case, rewrite and retry. "
            "If the query is correct, it will be parsed and returned. "
            f"ALWAYS run this tool before executing a query with {query_sql_database_tool.name}. "
        )
        mindsdb_sql_parser_tool = MindsDBSQLParserTool(
            name=f'mindsdb_sql_parser_tool{prefix}',
            description=mindsdb_sql_parser_tool_description
        )

        return [
            query_sql_database_tool,
            info_sql_database_tool,
            list_sql_database_tool,
            mindsdb_sql_parser_tool,
        ]



__title__ = "MindsDB PyPI handler"
__package_name__ = "mindsdb_pypi_handler"
__version__ = "0.1.0"
__description__ = "MindsDB handler for PyPI"
__author__ = "Sadra Yahyapour"
__github__ = "https://github.com/mindsdb/mindsdb"
__pypi__ = "https://pypi.org/project/mindsdb/"
__license__ = "MIT"
__copyright__ = "Copyright 2023 - mindsdb"



from mindsdb.integrations.handlers.postgres_handler.postgres_handler import PostgresHandler


class CockroachHandler(PostgresHandler):
    """
    This handler handles connection and execution of the Cockroachdb statements.
    """
    name = 'cockroachdb'

    def __init__(self, name=None, **kwargs):
        super().__init__(name, **kwargs)



from mindsdb.interfaces.storage import db
from mindsdb.interfaces.storage.fs import RESOURCE_GROUP
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class JsonStorage:
    def __init__(self, resource_group: str, resource_id: int):
        self.resource_group = resource_group
        self.resource_id = resource_id

    def __setitem__(self, key, value):
        if isinstance(value, dict) is False:
            raise TypeError(f"got {type(value)} instead of dict")
        existing_record = self.get_record(key)
        if existing_record is None:
            record = db.JsonStorage(
                name=key,
                resource_group=self.resource_group,
                resource_id=self.resource_id,
                company_id=ctx.company_id,
                content=value
            )
            db.session.add(record)
        else:
            existing_record.content = value
        db.session.commit()

    def set(self, key, value):
        self[key] = value

    def __getitem__(self, key):
        record = self.get_record(key)
        if record is None:
            return None
        return record.content

    def get(self, key):
        return self[key]

    def get_record(self, key):
        record = db.session.query(db.JsonStorage).filter_by(
            name=key,
            resource_group=self.resource_group,
            resource_id=self.resource_id,
            company_id=ctx.company_id
        ).first()
        return record

    def get_all_records(self):
        records = db.session.query(db.JsonStorage).filter_by(
            resource_group=self.resource_group,
            resource_id=self.resource_id,
            company_id=ctx.company_id
        ).all()
        return records

    def __repr__(self):
        records = self.get_all_records()
        names = [x.name for x in records]
        return f'json_storage({names})'

    def __len__(self):
        records = self.get_all_records()
        return len(records)

    def __delitem__(self, key):
        record = self.get_record(key)
        if record is not None:
            try:
                db.session.delete(record)
                db.session.commit()
            except Exception:
                db.session.rollback()
                logger.error('cant delete record from JSON storage')

    def delete(self, key):
        del self[key]

    def clean(self):
        json_records = self.get_all_records()
        for record in json_records:
            db.session.delete(record)
        try:
            db.session.commit()
        except Exception:
            db.session.rollback()
            logger.error('cant delete records from JSON storage')


def get_json_storage(resource_id: int, resource_group: str = RESOURCE_GROUP.PREDICTOR):
    return JsonStorage(
        resource_group=resource_group,
        resource_id=resource_id,
    )



from typing import Any, Dict, List, Union
import logging
from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.messages.base import BaseMessage
from langchain_core.outputs import LLMResult


class ContextCaptureCallback(BaseCallbackHandler):
    def __init__(self):
        self.context = None

    def on_retriever_end(self, documents: List[Any], *, run_id: str, parent_run_id: Union[str, None] = None, **kwargs: Any) -> Any:
        self.context = [{
            'page_content': doc.page_content,
            'metadata': doc.metadata
        } for doc in documents]

    def get_contexts(self):
        return self.context


class LogCallbackHandler(BaseCallbackHandler):
    '''Langchain callback handler that logs agent and chain executions.'''

    def __init__(self, logger: logging.Logger):
        logger.setLevel('DEBUG')
        self.logger = logger
        self._num_running_chains = 0
        self.generated_sql = None

    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> Any:
        '''Run when LLM starts running.'''
        self.logger.debug('LLM started with prompts:')
        for prompt in prompts:
            self.logger.debug(prompt[:50])

    def on_chat_model_start(
            self,
            serialized: Dict[str, Any],
            messages: List[List[BaseMessage]], **kwargs: Any
    ) -> Any:
        '''Run when Chat Model starts running.'''
        self.logger.debug('Chat model started with messages:')
        for message_list in messages:
            for message in message_list:
                self.logger.debug(message.pretty_print())

    def on_llm_new_token(self, token: str, **kwargs: Any) -> Any:
        '''Run on new LLM token. Only available when streaming is enabled.'''
        pass

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> Any:
        '''Run when LLM ends running.'''
        self.logger.debug('LLM ended with response:')
        self.logger.debug(str(response.llm_output))

    def on_llm_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        '''Run when LLM errors.'''
        self.logger.debug(f'LLM encountered an error: {str(error)}')

    def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> Any:
        '''Run when chain starts running.'''
        self._num_running_chains += 1
        self.logger.info('Entering new LLM chain ({} total)'.format(
            self._num_running_chains))
        self.logger.debug('Inputs: {}'.format(inputs))

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> Any:
        '''Run when chain ends running.'''
        self._num_running_chains -= 1
        self.logger.info('Ended LLM chain ({} total)'.format(
            self._num_running_chains))
        self.logger.debug('Outputs: {}'.format(outputs))

    def on_chain_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        '''Run when chain errors.'''
        self._num_running_chains -= 1
        self.logger.error(
            'LLM chain encountered an error ({} running): {}'.format(
                self._num_running_chains, error))

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> Any:
        '''Run when tool starts running.'''
        pass

    def on_tool_end(self, output: str, **kwargs: Any) -> Any:
        '''Run when tool ends running.'''
        pass

    def on_tool_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        '''Run when tool errors.'''
        pass

    def on_text(self, text: str, **kwargs: Any) -> Any:
        '''Run on arbitrary text.'''
        pass

    def on_agent_action(self, action: AgentAction, **kwargs: Any) -> Any:
        '''Run on agent action.'''
        self.logger.debug(f'Running tool {action.tool} with input:')
        self.logger.debug(action.tool_input)

        stop_block = 'Observation: '
        if stop_block in action.tool_input:
            action.tool_input = action.tool_input[: action.tool_input.find(stop_block)]

        if action.tool.startswith("sql_db_query"):
            # Save the generated SQL query
            self.generated_sql = action.tool_input

        # fix for mistral
        action.tool = action.tool.replace('\\', '')

    def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> Any:
        '''Run on agent end.'''
        self.logger.debug('Agent finished with return values:')
        self.logger.debug(str(finish.return_values))



from enum import Enum


class PostgresBackendMessageIdentifier(Enum):
    NOTICE_RESPONSE = b'N'
    AUTHENTICATION_REQUEST = b'R'
    READY_FOR_QUERY = b'Z'
    COMPLETE = b'C'
    ERROR = b'E'
    ROW_DESCRIPTION = b'T'
    DATA_ROW = b'D'
    NEGOTIATE_VERSION = b'v'
    PARAMETER = b'S'
    PARSE_COMPLETE = b'1'
    BIND_COMPLETE = b'2'
    PARAMETER_DESCRIPTION = b't'


class PostgresFrontendMessageIdentifier(Enum):
    EXECUTE = b'E'
    QUERY = b'Q'
    TERMINATE = b'X'
    SSL_REQUEST = b'F'
    PARSE = b'P'
    BIND = b'B'
    SYNC = b'S'
    DESCRIBE = b'D'


class PostgresAuthType(Enum):
    PASSWORD = b'p'



from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'getParameter': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):

        if query['featureCompatibilityVersion']:
            res = {
                "featureCompatibilityVersion":
                    {
                        "version": "3.6"
                    },
                "ok": 1.0
            }
        else:
            raise NotImplementedError(f'Unknown parameter {query}')
        return res


responder = Responce()



import secrets
import threading
import time

from mindsdb_sql_parser.ast import Identifier, Select, Insert

from mindsdb.utilities import log
from mindsdb.utilities.context import context as ctx

from .types import ChatBotMessage, BotException

logger = log.getLogger(__name__)


class BasePolling:
    def __init__(self, chat_task, chat_params):
        self.params = chat_params
        self.chat_task = chat_task

    def start(self, stop_event):
        raise NotImplementedError

    def send_message(self, message: ChatBotMessage, table_name=None):
        chat_id = message.destination if isinstance(message.destination, tuple) else (message.destination,)
        text = message.text

        t_params = self.params["chat_table"] if table_name is None else next(
            (t["chat_table"] for t in self.params if t["chat_table"]["name"] == table_name)
        )
        chat_id_cols = t_params["chat_id_col"] if isinstance(t_params["chat_id_col"], list) else [t_params["chat_id_col"]]

        ast_query = Insert(
            table=Identifier(t_params["name"]),
            columns=[*chat_id_cols, t_params["text_col"]],
            values=[
                [*chat_id, text],
            ],
        )

        self.chat_task.chat_handler.query(ast_query)


class MessageCountPolling(BasePolling):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._to_stop = False
        self.chats_prev = None

    def run(self, stop_event):
        while True:
            try:
                for chat_params in self.params:
                    chat_ids = self.check_message_count(chat_params)
                    logger.debug(f"number of chat ids found: {len(chat_ids)}")

                    for chat_id in chat_ids:
                        try:
                            chat_memory = self.chat_task.memory.get_chat(
                                chat_id,
                                table_name=chat_params["chat_table"]["name"],
                            )
                        except Exception as e:
                            logger.error(f"Problem retrieving chat memory: {e}")

                        try:
                            message = self.get_last_message(chat_memory)
                        except Exception as e:
                            logger.error(f"Problem getting last message: {e}")
                            message = None

                        if message:
                            self.chat_task.on_message(message, chat_memory=chat_memory, table_name=chat_params["chat_table"]["name"])

            except Exception as e:
                logger.error(e)

            if stop_event.is_set():
                return
            logger.debug(f"running {self.chat_task.bot_id}")
            time.sleep(7)

    def get_last_message(self, chat_memory):
        # retrive from history
        try:
            history = chat_memory.get_history()
        except Exception as e:
            logger.error(f"Problem retrieving history: {e}")
            history = []
        last_message = history[-1]
        if last_message.user == self.chat_task.bot_params["bot_username"]:
            # the last message is from bot
            return
        return last_message

    def check_message_count(self, chat_params):
        p_params = chat_params["polling"]

        chat_ids = []

        id_cols = p_params["chat_id_col"] if isinstance(p_params["chat_id_col"], list) else [p_params["chat_id_col"]]
        msgs_col = p_params["count_col"]
        # get chats status info
        ast_query = Select(
            targets=[*[Identifier(id_col) for id_col in id_cols], Identifier(msgs_col)],
            from_table=Identifier(p_params["table"]),
        )

        resp = self.chat_task.chat_handler.query(query=ast_query)
        if resp.data_frame is None:
            raise BotException("Error to get count of messages")

        chats = {}
        for row in resp.data_frame.to_dict("records"):
            chat_id = tuple(row[id_col] for id_col in id_cols)
            msgs = row[msgs_col]

            chats[chat_id] = msgs

        if self.chats_prev is None:
            # first run
            self.chats_prev = chats
        else:
            # compare
            # for new keys
            for chat_id, count_msgs in chats.items():
                if self.chats_prev.get(chat_id) != count_msgs:
                    chat_ids.append(chat_id)

            self.chats_prev = chats
        return chat_ids

    def stop(self):
        self._to_stop = True


class RealtimePolling(BasePolling):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # call back can be without context
        self._ctx_dump = ctx.dump()

    def _callback(self, row, key):
        ctx.load(self._ctx_dump)

        row.update(key)

        # If more than one set of parameters is present, multiple tables are supported.
        if len(self.params) > 1:
            # Identify the table relevant to this event based on the key.
            event_keys = list(key.keys())
            for param in self.params:
                table_keys = [param["chat_table"]["chat_id_col"]] if isinstance(param["chat_table"]["chat_id_col"], str) else param["chat_table"]["chat_id_col"]

                if sorted(event_keys) == sorted(table_keys):
                    t_params = param["chat_table"]
                    break

        # Otherwise, only a single table is supported. Use the first set of parameters.
        else:
            t_params = self.params[0]

        # Get the chat ID from the row based on the chat ID column(s).
        chat_id = tuple(row[key] for key in t_params["chat_id_col"]) if isinstance(t_params["chat_id_col"], list) else row[t_params["chat_id_col"]]

        message = ChatBotMessage(
            ChatBotMessage.Type.DIRECT,
            row[t_params["text_col"]],
            # In Slack direct messages are treated as channels themselves.
            row[t_params["username_col"]],
            chat_id,
        )

        self.chat_task.on_message(
            message,
            chat_id=chat_id,
            table_name=t_params["name"],
        )

    def run(self, stop_event):
        self.chat_task.chat_handler.subscribe(
            stop_event,
            self._callback
        )

    # def send_message(self, message: ChatBotMessage):
    #
    #     self.chat_task.chat_handler.realtime_send(message)


class WebhookPolling(BasePolling):
    """
    Polling class for handling webhooks.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def run(self, stop_event: threading.Event) -> None:
        """
        Run the webhook polling.
        Check if a webhook token is set for the chatbot. If not, generate a new one.
        Then, do nothing, as the webhook is handled by a task instantiated for each request.

        Args:
            stop_event (threading.Event): Event to stop the polling.
        """
        # If a webhook token is not set for the chatbot, generate a new one.
        from mindsdb.interfaces.chatbot.chatbot_controller import ChatBotController

        chat_bot_controller = ChatBotController()
        chat_bot = chat_bot_controller.get_chatbot_by_id(self.chat_task.object_id)

        if not chat_bot["webhook_token"]:
            chat_bot_controller.update_chatbot(
                chatbot_name=chat_bot["name"],
                project_name=chat_bot["project"],
                webhook_token=secrets.token_urlsafe(32),
            )

        # Do nothing, as the webhook is handled by a task instantiated for each request.
        stop_event.wait()

    def send_message(self, message: ChatBotMessage, table_name: str = None) -> None:
        """
        Send a message (response) to the chatbot.
        Pass the message to the chatbot handler to respond.

        Args:
            message (ChatBotMessage): The message to send.
            table_name (str): The name of the table to send the message to. Defaults to None.
        """
        self.chat_task.chat_handler.respond(message)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the SQreamDB server/database.'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the SQreamDB server.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the SQreamDB server.',
        'secret': True
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect SQreamDB server'
    }, 
    database={
        'type': ARG_TYPE.STR,
        'description': 'Specify database name  to connect SQreamDB server'
    },
    service={
        'type': ARG_TYPE.STR,
        'description': 'Optional: service queue (default: "sqream")'
    },
    use_ssl={
        'type': ARG_TYPE.BOOL,
        'description': 'use SSL connection (default: False)'
    },
    clustered={
        'type': ARG_TYPE.BOOL,
        'description': 'Optional:  Connect through load balancer, or direct to worker (Default: false - direct to worker)'
    },
)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=5000,
    password='sqream',
    user='master',
    database='sqream'
)



import time
from typing import Dict
from http import HTTPStatus
from sqlalchemy.exc import NoResultFound

from flask import request
from flask_restx import Resource

from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.databases import ns_conf
from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.api.executor.datahub.classes.tables_row import TablesRow
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb_sql_parser import parse_sql, ParsingException
from mindsdb_sql_parser.ast import CreateTable, DropTables
from mindsdb.utilities.exception import EntityNotExistsError
from mindsdb.integrations.libs.response import HandlerStatusResponse


@ns_conf.route('/')
class DatabasesResource(Resource):
    @ns_conf.doc('list_databases')
    @api_endpoint_metrics('GET', '/databases')
    def get(self):
        '''List all databases'''
        session = SessionController()
        return session.database_controller.get_list()

    @ns_conf.doc('create_database')
    @api_endpoint_metrics('POST', '/databases')
    def post(self):
        '''Create a database'''
        if 'database' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Wrong argument',
                'Must provide "database" parameter in POST body'
            )
        check_connection = request.json.get('check_connection', False)
        session = SessionController()
        database = request.json['database']
        parameters = {}
        if 'name' not in database:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Wrong argument',
                'Missing "name" field for database'
            )
        if 'engine' not in database:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Wrong argument',
                'Missing "engine" field for database. If you want to create a project instead, use the /api/projects endpoint.'
            )
        if 'parameters' in database:
            parameters = database['parameters']
        name = database['name']

        if session.database_controller.exists(name):
            return http_error(
                HTTPStatus.CONFLICT, 'Name conflict',
                f'Database with name {name} already exists.'
            )

        storage = None
        if check_connection:
            try:
                handler = session.integration_controller.create_tmp_handler(name, database['engine'], parameters)
                status = handler.check_connection()
            except ImportError as import_error:
                status = HandlerStatusResponse(success=False, error_message=str(import_error))

            if status.success is not True:
                if hasattr(status, 'redirect_url') and isinstance(status, str):
                    return {
                        "status": "redirect_required",
                        "redirect_url": status.redirect_url,
                        "detail": status.error_message
                    }, HTTPStatus.OK
                return {
                    "status": "connection_error",
                    "detail": status.error_message
                }, HTTPStatus.OK

            if status.copy_storage:
                storage = handler.handler_storage.export_files()

        new_integration_id = session.integration_controller.add(name, database['engine'], parameters)

        if storage:
            handler = session.integration_controller.get_data_handler(name, connect=False)
            handler.handler_storage.import_files(storage)

        new_integration = session.database_controller.get_integration(new_integration_id)
        return new_integration, HTTPStatus.CREATED


@ns_conf.route('/<database_name>')
class DatabaseResource(Resource):
    @ns_conf.doc('get_database')
    @api_endpoint_metrics('GET', '/databases/database')
    def get(self, database_name):
        '''Gets a database by name'''
        session = SessionController()
        check_connection = request.args.get('check_connection', 'false').lower() in ('1', 'true')
        try:
            project = session.database_controller.get_project(database_name)
            result = {
                'name': database_name,
                'type': 'project',
                'id': project.id,
                'engine': None
            }
            if check_connection:
                result['connection_status'] = {
                    'success': True,
                    'error_message': None
                }
        except (ValueError, EntityNotExistsError):
            integration = session.integration_controller.get(database_name)
            if integration is None:
                return http_error(
                    HTTPStatus.NOT_FOUND, 'Database not found',
                    f'Database with name {database_name} does not exist.'
                )
            result = integration
            if check_connection:
                integration['connection_status'] = {
                    'success': False,
                    'error_message': None
                }
                try:
                    handler = session.integration_controller.get_data_handler(database_name)
                    status = handler.check_connection()
                    integration['connection_status']['success'] = status.success
                    integration['connection_status']['error_message'] = status.error_message
                except Exception as e:
                    integration['connection_status']['success'] = False
                    integration['connection_status']['error_message'] = str(e)

        return result

    @ns_conf.doc('update_database')
    @api_endpoint_metrics('PUT', '/databases/database')
    def put(self, database_name):
        '''Updates or creates a database'''
        if 'database' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Wrong argument',
                'Must provide "database" parameter in POST body'
            )

        session = SessionController()
        parameters = {}
        database = request.json['database']
        check_connection = request.json.get('check_connection', False)

        if 'parameters' in database:
            parameters = database['parameters']
        if not session.database_controller.exists(database_name):
            # Create.
            if 'engine' not in database:
                return http_error(
                    HTTPStatus.BAD_REQUEST, 'Wrong argument',
                    'Missing "engine" field for new database. '
                    'If you want to create a project instead, use the POST /api/projects endpoint.'
                )
            new_integration_id = session.integration_controller.add(database_name, database['engine'], parameters)
            new_integration = session.database_controller.get_integration(new_integration_id)
            return new_integration, HTTPStatus.CREATED

        if check_connection:
            existing_integration = session.integration_controller.get(database_name)
            temp_name = f'{database_name}_{time.time()}'.replace('.', '')
            try:
                handler = session.integration_controller.create_tmp_handler(
                    temp_name, existing_integration['engine'], parameters
                )
                status = handler.check_connection()
            except ImportError as import_error:
                status = HandlerStatusResponse(success=False, error_message=str(import_error))

            if status.success is not True:
                return http_error(
                    HTTPStatus.BAD_REQUEST, 'Connection error',
                    status.error_message or 'Connection error'
                )

        session.integration_controller.modify(database_name, parameters)
        return session.integration_controller.get(database_name)

    @ns_conf.doc('delete_database')
    @api_endpoint_metrics('DELETE', '/databases/database')
    def delete(self, database_name):
        '''Deletes a database by name'''
        session = SessionController()
        if not session.database_controller.exists(database_name):
            return http_error(
                HTTPStatus.NOT_FOUND, 'Database not found',
                f'Database with name {database_name} does not exist.'
            )
        try:
            session.database_controller.delete(database_name)
        except Exception as e:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Error',
                f'Cannot delete database {database_name}. '
                + 'This is most likely a system database, a permanent integration, or an ML engine with active models. '
                + f'Full error: {e}. '
                + 'Please check the name and try again.'
            )
        return '', HTTPStatus.NO_CONTENT


def _tables_row_to_obj(table_row: TablesRow) -> Dict:
    type = table_row.TABLE_TYPE.lower()
    if table_row.TABLE_TYPE == 'BASE TABLE':
        type = 'data'
    return {
        'name': table_row.TABLE_NAME,
        'type': type
    }


@ns_conf.route('/<database_name>/tables')
class TablesList(Resource):
    @ns_conf.doc('list_tables')
    @api_endpoint_metrics('GET', '/databases/database/tables')
    def get(self, database_name):
        '''Get all tables in a database'''
        session = SessionController()
        datanode = session.datahub.get(database_name)
        all_tables = datanode.get_tables()
        table_objs = [_tables_row_to_obj(t) for t in all_tables]
        return table_objs

    @ns_conf.doc('create_table')
    @api_endpoint_metrics('POST', '/databases/database/tables')
    def post(self, database_name):
        '''Creates a table in a database'''
        if 'table' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Wrong argument',
                'Must provide "table" parameter in POST body'
            )
        table = request.json['table']
        if 'name' not in table:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Wrong argument',
                'Missing "name" field for table'
            )
        if 'select' not in table:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Wrong argument',
                'Missing "select" field for table'
            )
        table_name = table['name']
        select_query = table['select']
        replace = False
        if 'replace' in table:
            replace = table['replace']

        session = SessionController()
        try:
            session.database_controller.get_project(database_name)
            error_message = f'Database {database_name} is a project. ' \
                + f'If you want to create a model or view, use the projects/{database_name}/models/{table_name} or ' \
                + f'projects/{database_name}/views/{table_name} endpoints instead.'
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Error',
                error_message
            )
        except NoResultFound:
            # Only support creating tables from integrations.
            pass

        datanode = session.datahub.get(database_name)
        if datanode is None:
            return http_error(
                HTTPStatus.NOT_FOUND, 'Database not exists',
                f'Database with name {database_name} does not exist'
            )
        all_tables = datanode.get_tables()
        for t in all_tables:
            if t.TABLE_NAME == table_name and not replace:
                return http_error(
                    HTTPStatus.CONFLICT, 'Name conflict',
                    f'Table with name {table_name} already exists'
                )

        try:
            select_ast = parse_sql(select_query)
        except ParsingException:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Error',
                f'Could not parse select statement {select_query}'
            )

        create_ast = CreateTable(
            f'{database_name}.{table_name}',
            from_select=select_ast,
            is_replace=replace
        )

        mysql_proxy = FakeMysqlProxy()

        try:
            mysql_proxy.process_query(create_ast.get_string())
        except Exception as e:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Error',
                str(e)
            )

        all_tables = datanode.get_tables()
        try:
            matching_table = next(t for t in all_tables if t.TABLE_NAME == table_name)
            return _tables_row_to_obj(matching_table), HTTPStatus.CREATED
        except StopIteration:
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR, 'Error',
                f'Table with name {table_name} could not be created'
            )


@ns_conf.route('/<database_name>/tables/<table_name>')
@ns_conf.param('database_name', 'Name of the database')
@ns_conf.param('table_name', 'Name of the table')
class TableResource(Resource):
    @ns_conf.doc('get_table')
    @api_endpoint_metrics('GET', '/databases/database/tables/table')
    def get(self, database_name, table_name):
        session = SessionController()
        datanode = session.datahub.get(database_name)
        all_tables = datanode.get_tables()
        try:
            matching_table = next(t for t in all_tables if t.TABLE_NAME == table_name)
            return _tables_row_to_obj(matching_table)
        except StopIteration:
            return http_error(
                HTTPStatus.NOT_FOUND, 'Table not found',
                f'Table with name {table_name} not found'
            )

    @ns_conf.doc('drop_table')
    @api_endpoint_metrics('DELETE', '/databases/database/tables/table')
    def delete(self, database_name, table_name):
        session = SessionController()
        try:
            session.database_controller.get_project(database_name)
            error_message = f'Database {database_name} is a project. ' \
                + f'If you want to delete a model or view, use the projects/{database_name}/models/{table_name} or ' \
                + f'projects/{database_name}/views/{table_name} endpoints instead.'
            return http_error(HTTPStatus.BAD_REQUEST, 'Error', error_message)
        except NoResultFound:
            # Only support dropping tables from integrations.
            pass

        datanode = session.datahub.get(database_name)
        if datanode is None:
            return http_error(
                HTTPStatus.NOT_FOUND, 'Database not found',
                f'Database with name {database_name} not found'
            )
        all_tables = datanode.get_tables()
        try:
            next(t for t in all_tables if t.TABLE_NAME == table_name)
        except StopIteration:
            return http_error(
                HTTPStatus.NOT_FOUND, 'Table not found',
                f'Table with name {table_name} not found'
            )

        drop_ast = DropTables(
            tables=[table_name],
            if_exists=True
        )

        try:
            integration_handler = session.integration_controller.get_data_handler(database_name)
        except Exception:
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR, 'Error',
                f'Could not get database handler for {database_name}'
            )
        try:
            result = integration_handler.query(drop_ast)
        except NotImplementedError:
            return http_error(
                HTTPStatus.BAD_REQUEST, 'Error',
                f'Database {database_name} does not support dropping tables.'
            )
        if result.type == RESPONSE_TYPE.ERROR:
            return http_error(HTTPStatus.BAD_REQUEST, 'Error', result.error_message)
        return '', HTTPStatus.NO_CONTENT



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    credentials={
        'type': ARG_TYPE.PATH,
        'description': 'Service Account Keys',
        'label': 'Upload Service Account Keys',
    },
)



import os

from eventbrite.client import Client
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse

from .eventbrite_tables import EventbriteUserTable
from .eventbrite_tables import EventbriteOrganizationTable
from .eventbrite_tables import EventbriteCategoryTable
from .eventbrite_tables import EventbriteSubcategoryTable
from .eventbrite_tables import EventbriteFormatTable
from .eventbrite_tables import EventbriteEventDetailsTable
from .eventbrite_tables import EventbriteEventsTable

logger = log.getLogger(__name__)


class EventbriteHandler(APIHandler):
    """A class for handling connections and interactions with the Eventbrite API.

    Attributes:
        api (Client): The `Client` object for accessing Eventbrite API.
    """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)

        args = kwargs.get("connection_data", {})
        self.connection_args = {}
        handler_config = Config().get("eventbrite_handler", {})

        # Set up connection arguments
        for k in ["access_token"]:
            if k in args:
                self.connection_args[k] = args[k]
            elif f"EVENTBRITE_{k.upper()}" in os.environ:
                self.connection_args[k] = os.environ[f"EVENTBRITE_{k.upper()}"]
            elif k in handler_config:
                self.connection_args[k] = handler_config[k]

        self.api = None
        self.is_connected = False

        userTable = EventbriteUserTable(self)
        self._register_table("user", userTable)

        organizationTable = EventbriteOrganizationTable(self)
        self._register_table("organization", organizationTable)

        categoryTable = EventbriteCategoryTable(self)
        self._register_table("category", categoryTable)

        subcategoryTable = EventbriteSubcategoryTable(self)
        self._register_table("subcategory", subcategoryTable)

        formatTable = EventbriteFormatTable(self)
        self._register_table("formats", formatTable)

        eventDetailsTable = EventbriteEventDetailsTable(self)
        self._register_table("event_details", eventDetailsTable)

        eventsTable = EventbriteEventsTable(self)
        self._register_table("events", eventsTable)

    def connect(self):
        """Initialize the Eventbrite API Client."""
        if self.is_connected:
            return self.api

        self.api = Client(access_token=self.connection_args["access_token"])

        self.is_connected = True
        return self.api

    def check_connection(self) -> StatusResponse:
        """Check the connection to the Eventbrite API."""
        response = StatusResponse(False)

        try:
            api = self.connect()
            me = api.get_current_user()
            logger.info(f"Connected to Eventbrite as {me['name']}")
            response.success = True

        except Exception as e:
            response.error_message = f"Error connecting to Eventbrite API: {e}"
            logger.error(response.error_message)

        if not response.success and self.is_connected:
            self.is_connected = False

        return response
